if USPF == nil then USPF = {} end
USPF.AddonName = "USPF"
USPF.version = 1.0
USPF.active = false
USPF.cache = {}
USPF.GUI = {}
local selectedChar = GetCurrentCharacterId()
local currentCharName = nil
local GZNBId, GCCId, GCQI = GetZoneNameById, GetCurrentCharacterId, GetCompletedQuestInfo
local IAchC, GAchNCr, GAchCr = IsAchievementComplete, GetAchievementNumCriteria, GetAchievementCriterion
local GS, zf, strF = GetString, zo_strformat, string.format

local USPF_LTF = LibTableFunctions
if not USPF_LTF then return end

USPF.settings = {
	title = {font = "ProseAntique",},
	GSP = {
		font = "Univers 57",
		doneColor = {1, 1, 1},
		needColor = {1, 1, 1},
		progColor = {1, 1, 1},
	},
	SQS = {
		font = "Univers 57",
		doneColorSS = {1, 1, 1},
		doneColorZQ = {1, 1, 1},
		needColorSS = {1, 0, 0},
		needColorZQ = {1, 0, 0},
		progColorSS = {0.7843, 0.3922, 0},
		progColorZQ = {0.7843, 0.3922, 0},
		sortCol = 1,
	},
	GDQ = {
		font = "Univers 57",
		doneColor = {1, 1, 1},
		needColor = {1, 0, 0},
		sortCol = 1,
	},
	PDB = {
		font = "Univers 57",
		doneColor = {1, 1, 1},
		needColor = {1, 0, 0},
		sortCol = 1,
	},
	FD = {
		override = false,
		charHasFD = false,
	},
	MWC = false,
	SSC = false,
	EWC = false,
	GMC = false,
	BWC = false,
}

USPF.defaults = {
	firstRun = true,
	numChars = 0,
	charInfo = {},
	settings = {},
	ptsData = {},
}


USPF.ptsData = {
	Tot	= 0, GenTot	= 0, ZQTot	= 0, numSSTot	= 0, SSTot	= 0,
	GDTot	= 0, PDTot	= 0, Level	= 0, MainQ	= 0, FolDis	= 0,
	MWChar	= 0, SUChar	= 0, EWChar	= 0, GMChar	= 0, BWChar = 0,
	PvPRank	= 0, MaelAr	= 0, Unassigned = nil,
	ZQ = {
		AD0 = 0, AD1 = 0, AD2 = 0, AD3 = 0, AD4 = 0, AD5  = 0, DC0a = 0, DC0b = 0,
		DC1 = 0, DC2 = 0, DC3 = 0, DC4 = 0, DC5 = 0, EP0a = 0, EP0b = 0, EP1  = 0,
		EP2 = 0, EP3 = 0, EP4 = 0, EP5 = 0, CH  = 0, CAD  = 0, CDC  = 0, CEP  = 0,
		CMT = 0, LCL = 0, UCL = 0, CC  = 0, DB  = 0, IC   = 0, MW   = 0, RO   = 0,
		TG  = 0, SU  = 0, MM  = 0, NE  = 0, WP  = 0, SE   = 0, WS   = 0, TR   = 0,
		BW  = 0
	},
	SS = {
		AD0 = 0, AD1 = 0, AD2 = 0, AD3 = 0, AD4 = 0, AD5  = 0, DC0a = 0, DC0b = 0,
		DC1 = 0, DC2 = 0, DC3 = 0, DC4 = 0, DC5 = 0, EP0a = 0, EP0b = 0, EP1  = 0,
		EP2 = 0, EP3 = 0, EP4 = 0, EP5 = 0, CH  = 0, CAD  = 0, CDC  = 0, CEP  = 0,
		CMT = 0, LCL = 0, UCL = 0, IC  = 0, WR  = 0, HB   = 0, GC   = 0, VV   = 0,
		CC  = 0, MQ  = 0, SU  = 0, MM  = 0, NE  = 0, SE   = 0, WS   = 0, TR   = 0,
		BW  = 0
	},
	GD = {
		BC1 = 0, BC2 = 0, EH1 = 0, EH2 = 0, CA1 = 0, CA2 = 0, TI = 0, SW = 0,
		SC1 = 0, SC2 = 0, WS1 = 0, WS2 = 0, CH1 = 0, CH2 = 0, VF = 0, BH = 0,
		FG1 = 0, FG2 = 0, DC1 = 0, DC2 = 0, AC  = 0, DK  = 0, BC = 0, VM = 0,
		WGT = 0, ICP = 0, RM  = 0, CS  = 0, BF  = 0, FH  = 0, FL = 0, SP = 0,
		MHK = 0, MOS = 0, DoM = 0, FV  = 0, LM  = 0, MF  = 0, IR = 0, UG = 0,
		SG  = 0, CT  = 0, BDV = 0, TC  = 0, RPB = 0, TDC = 0,
	},
	PD = {
		AD1 = 0, AD2 = 0, AD3 = 0, AD4 = 0, AD5 = 0, DC1 = 0, DC2 = 0, DC3 = 0,
		DC4 = 0, DC5 = 0, EP1 = 0, EP2 = 0, EP3 = 0, EP4 = 0, EP5 = 0, CH  = 0,
		VFW = 0, VNC = 0, WOO = 0, WRK = 0, SKW = 0, SSH = 0, RN  = 0, OC  = 0,
		LT  = 0, NK  = 0, SH  = 0, ZA  = 0
	},
}

USPF.ptsTots = {	--Tot and GenTot are 481 and 129 because you can't do more than one DLC tutorial.
	Tot		= 483, GenTot = 129, ZQTot	= 115, numSSTot	= 495, SSTot  = 165,
	GDTot	=  46, PDTot  =  28, Level	=  64, MainQ	=  11, FolDis =   2,
	MWChar	=   1, SUChar =   1, EWChar	=   1, GMChar	=   1, BWChar =   1,
	PvPRank	=  50, MaelAr =   1,
	ZQ = {
		AD0 = 0, AD1 = 3, AD2 =  3, AD3 = 3, AD4 = 3, AD5  = 3, DC0a = 0, DC0b = 0,
		DC1 = 4, DC2 = 3, DC3 =  3, DC4 = 3, DC5 = 3, EP0a = 0, EP0b = 0, EP1  = 3,
		EP2 = 3, EP3 = 3, EP4 =  3, EP5 = 3, CH  = 3, CAD  = 0, CDC  = 0, CEP  = 0,
		CMT = 0, LCL = 0, UCL =  0, CC  = 8, DB  = 8, IC   = 1, MW   = 3, RO   = 3,
		TG  = 6, SU  = 3, MM  =  7, NE  = 3, WP  = 0, SE   = 9, WS   = 3, TR   = 9,
		BW  = 3
	},
	SS = {
		AD0  = 6,  AD1 = 16, AD2 = 16, AD3 = 16, AD4 = 16, AD5 = 16, DC0a = 3,
		DC0b = 3,  DC1 = 16, DC2 = 16, DC3 = 16, DC4 = 16, DC5 = 16, EP0a = 3,
		EP0b = 3,  EP1 = 16, EP2 = 16, EP3 = 16, EP4 = 16, EP5 = 16, CH   = 16,
		CAD  = 15, CDC = 15, CEP = 15, CMT = 1,  LCL = 12, UCL = 6,  IC   = 13,
		WR   = 17, HB  = 6,  GC  = 6,  VV  = 18, CC  = 6,  MQ  = 1,  SU   = 18,
		MM   = 6,  NE  = 18, SE  = 6,  WS  = 18, TR  = 6,  BW  = 18
	},
	GD = {
		BC1 = 1, BC2 = 1, EH1 = 1, EH2 = 1, CA1 = 1, CA2 = 1, TI = 1, SW = 1,
		SC1 = 1, SC2 = 1, WS1 = 1, WS2 = 1, CH1 = 1, CH2 = 1, VF = 1, BH = 1,
		FG1 = 1, FG2 = 1, DC1 = 1, DC2 = 1, AC  = 1, DK  = 1, BC = 1, VM = 1,
		WGT = 1, ICP = 1, RM  = 1, CS  = 1, BF  = 1, FH  = 1, FL = 1, SP = 1,
		MHK = 1, MOS = 1, DoM = 1, FV  = 1, LM  = 1, MF  = 1, IR = 1, UG = 1,
		SG  = 1, CT  = 1, BDV = 1, TC  = 1, RPB = 1, TDC = 1,
	},
	PD = {
		AD1 = 1, AD2 = 1, AD3 = 1, AD4 = 1, AD5 = 1, DC1 = 1, DC2 = 1, DC3 = 1,
		DC4 = 1, DC5 = 1, EP1 = 1, EP2 = 1, EP3 = 1, EP4 = 1, EP5 = 1, CH  = 1,
		VFW = 1, VNC = 1, WOO = 1, WRK = 1, SKW = 1, SSH = 1, RN  = 1, OC  = 1,
		LT  = 1, NK  = 1, SH  = 1, ZA  = 1
	},
}

local tempZId = {
	ZN = {
		AD0  =  537, AD1  =  381, AD2  = 383, AD3 =  108, AD4 =  58, AD5 =  382,
		DC0A =  535, DC0B =  534, DC1  =   3, DC2 =   19, DC3 =  20, DC4 =  104,
		DC5  =   92, EP0B =  280, EP0A = 281, EP1 =   41, EP2 =  57, EP3 =  117,
		EP4  =  101, EP5  =  103, CH   = 347, CYD =  181, CAD = 181, CDC =  181,
		CEP  =  181, CMT  =  181, CL   = 888, LCL =  888, UCL = 888, IC  =  584,
		RO   =  684, HB   =  816, GC   = 823, VV  =  849, CC  = 980, SU  = 1011,
		MM   =  726, NE   = 1086, WP   = 586, SE  = 1133, WS = 1160, BGC = 1161,
		TR   = 1207, BW   = 1261,
	},
	GDN = {
		BC1 =  380, BC2 =  935, EH1 =  126, EH2 =  931, CA1 =  176, CA2 =  681,
		TI  =  131, SW  =   31, SC1 =  144, SC2 =  936, WS1 =  146, WS2 =  933,
		CH1 =  130, CH2 =  932, VF  =   22, BH  =   38, FG1 =  283, FG2 =  934,
		DC1 =   63, DC2 =  930, AC  =  148, DK  =  449, BC  =   64, VM  =   11,
		ICP =  678, WGT =  688, CS  =  848, RM  =  843, BF  =  973, FH  =  974,
		FL  = 1009, SP  = 1010, MHK = 1052, MOS = 1055, DoM = 1081, FV  = 1080,
		LM  = 1123, MF  = 1122, IR  = 1152, UG  = 1153, SG  = 1197, CT  = 1201,
		BDV = 1228, TC  = 1229, RPB = 1267, TDC = 1268,
	},
	PDN = {
		AD1 =  486, AD2 =  124, AD3 =  137, AD4 =  138, AD5 =  487, DC1 =  284,
		DC2 =  142, DC3 =  162, DC4 =  308, DC5 =  169, EP1 =  216, EP2 =  306,
		EP3 =  134, EP4 =  339, EP5 =  341, CH  =  557, WOO =  706, WRK =  705,
		VFW =  919, VNC =  918, SKW = 1020, SSH = 1021, RN  = 1089, OC  = 1090,
		LT  = 1186, NK  = 1187, SH  = 1260, ZA  = 1259,
	},
}

USPF.data = {
	ZId = tempZId,
	MAAch = 1304,
	MQAch = 1003,
	ZQAch = {
		AD1 =  943, AD2 =  944, AD3 =  945, AD4 =  946, AD5 =  947, CC  = 2064,
		CH  =  957, DB  = 1444, DC1 =  953, DC2 =  954, DC3 =  955, DC4 =  956,
		DC5 =  958, EP1 =  948, EP2 =  949, EP3 =  950, EP4 =  951, EP5 =  952,
		IC  = 1175, RO  = 1260, TG  = 1363, SU  = 2208, MM  = 2339, NE  = 2488,
		SE  = 2604, WS  = 2716, BW  = 3055
	},
	AD1 = {	--943
		{4222,  360},
		{4345,  361},
		{4261,  362},
	},
	AD2 = {	--944	(1-2 only)
		{4868,  605},
		{4386,  606},
		{4885,  607},
	},
	AD3 = {	--945
		{4750,  510},
		{4765,  511},
		{4690,  512},
	},
	AD4 = {	--946
		{4337,    0},
		{4452,  283},
		{4143,  285},
	},
	AD5 = {	--947
		{4712,    0},
		{4479,  537},
		{4720,  538},
	},
	CC = {	--2064
		{6050, 2050},
		{6057, 2058},
		{6063, 2059},
		{6025, 2060},
		{6052, 2061},
		{6046, 2062},
		{6047, 2063},
		{6048, 2064},
	},
	CH = {	--957
		{4602,  612},
		{4730,  613},
		{4758,  614},
	},
	DB = {	--1444
		{5540, 1436},
		{5595, 1438},
		{5599, 1439},
		{5596, 1440},
		{5567, 1441},
		{5597, 1442},
		{5598, 1443},
		{5600, 1444},
	},
	DC1 = {	--953
		{3006,    0},
		{3235,   30},
		{3267,   28},
		{3379,   31},
	},
	DC2 = {	--954
		{ 467,  154},
		{1633,  155},
		{ 575,    0},
	},
	DC3 = {	--955
		{ 465,  589},
		{4972,  590},
		{4884,  591},
	},
	DC4 = {	--956
		{2192,  516},
		{2222,  517},
		{2997,  956},
	},
	DC5 = {	--958
		{4891,    0},
		{4912,  146},
		{4960,  147},
	},
	EO = {
		{6324,    0},
	},
	EP1 = {	--948
		{3735,  201},
		{3634,  204},
		{3868,  209},
	},
	EP2 = {	--949
		{3797,  363},
		{3817,  364},
		{3831,  365},
	},
	EP3 = {	--950
		{4590,  184},
		{4606,  185},
		{3910,  186},
	},
	EP4 = {	--951
		{4061,    0},
		{4115,    0},
		{4117,  599},
	},
	EP5 = {	--952
		{3968,  335},
		{4139,  336},
		{4188,  337},
	},
	GD = {
		BC1 = 4107,
		BC2 = 4597,
		EH1 = 4336,
		EH2 = 4675,
		CA1 = 4778,
		CA2 = 5120,
		TI  = 4538,
		SW  = 4733,
		SC1 = 4054,
		SC2 = 4555,
		WS1 = 4246,
		WS2 = 4813,
		CH1 = 4379,
		CH2 = 5113,
		VF  = 4432,
		BH  = 4589,
		FG1 = 3993,
		FG2 = 4303,
		DC1 = 4145,
		DC2 = 4641,
		AC  = 4202,
		DK  = 4346,
		BC  = 4469,
		VM  = 4822,
		ICP = 5136,
		WGT = 5342,
		CS  = 5702,
		RM  = 5403,
		BF  = 5889,
		FH  = 5891,
		FL  = 6064,
		SP  = 6065,
		MHK = 6186,
		MOS = 6188,
		DoM = 6251,
		FV  = 6249,
		LM  = 6351,
		MF  = 6349,
		IR  = 6414,
		UG  = 6416,
		SG  = 6505,
		CT  = 6507,
		BDV = 6576,
		TC  = 6578,
		RPB = 6683,
		TDC = 6685,
	},
	IC = {	--1175
		{5482, 1175},
	},
	MM = {	--2339
		{6246, 2333},
		{6266, 2334},
		{6241, 2335},
		{6259, 2336},
		{6243, 2337},
		{6244, 2338},
		{6245, 2339},
	},
	MW = {
		{6003, 1852},
		{5922, 1869},
		{5948, 1870},
	},
	MO = {
		{5804,    0},
	},
	MQ = {	--1003
		{4296,  993},
		{4831,  994},
		{4474,  995},
		{4552,  996},
		{4607,  997},
		{4764,  998},
		{4836,  999},
		{4837, 1000},
		{4867, 1001},
		{4832, 1002},
		{4847, 1003},
	},
	NE = {	--2488
		{6336, 2482},
		{6304, 2485},
		{6315, 2488},
	},
	RO = {	--1260
		{5447, 1325},
		{5468, 1326},
		{5481, 1327},
	},
	TG = {	--1363
		{5531, 1371},
		{5534, 1360},
		{5532, 1370},
		{5556, 1361},
		{5549, 1362},
		{5545, 1363},
	},
	SE = {	--2604
		{6401, 2596},
		{6409,    0},
		{6394, 2597},
		{6399, 2598},
		{6403, 2599},
		{6404, 2600},
		{6393, 2602},
		{6397, 2603},
		{6402, 2604},
	},
	SO = {
		{6143,    0},
	},
	SU = {	--2208
		{6132, 2203},
		{6113, 2206},
		{6126, 2208},
	},
	WS = {	--2722
		{6476, 2720},
		{6466, 2722},
		{6481, 2725},
	},
	GO = {
		{6455,    0},
	},
	BO = {
		{6646,    0},
	},
	TR = {
		{6550, 2930},
		{6551, 2931},
		{6547, 2932},
		{6548, 2933},
		{6554, 2934},
		{6566, 2935},
		{6552, 2937},
		{6560, 2936},
		{6570, 2939},
	},
	BW = {
		{6616, 3049},
		{6619, 3051},
		{6660, 3054},
	},
	PD = {
		AD1	=  468,
		AD2	=  470,
		AD3	=  445,
		AD4	=  460,
		AD5	=  469,
		DC1	=  380,
		DC2	=  714,
		DC3	=  713,
		DC4	=  707,
		DC5	=  708,
		EP1	=  379,
		EP2	=  388,
		EP3	=  372,
		EP4	=  381,
		EP5	=  371,
		CH	=  874,
		VFW	= 1855,
		VNC	= 1846,
		WOO	= 1238,
		WRK	= 1235,
		SKW	= 2096,
		SSH	= 2095,
		RN	= 2444,
		OC	= 2445,
		LT	= 2714,
		NK	= 2715,
		SH	= 2994,
		ZA	= 2995,
	},
	SS = {
		MQ	 = 2521,
		AD0	 =  431,
		AD1	 =  695,
		AD2	 =  682,
		AD3	 =  683,
		AD4	 =  684,
		AD5	 =  685,
		DC0a =  408,
		DC0b =  407,
		DC1	 =  409,
		DC2	 =  515,
		DC3	 =  554,
		DC4	 =  556,
		DC5	 =  557,
		EP0a =  405,
		EP0b =  398,
		EP1	 =  397,
		EP2	 =  547,
		EP3	 =  687,
		EP4	 =  688,
		EP5	 =  689,
		CH	 =  686,
		CAD	 =  694,
		CDC	 =  693,
		CEP	 =  692,
		CMT	 =  748,
		LCL	 =  727,
		UCL	 =  912,
		IC	 = 1160,
		WR	 = 1320,
		HB	 = 1347,
		GC	 = 1342,
		VV	 = 1843,
		CC	 = 1844,
		SU	 = 1845,
		MM	 = 2291,
		NE	 = 2461,
		SE	 = 2562,
		WS	 = 2687,
		TR	 = 2857,
		BW	 = 2982,
	},
	racialLineIds = {
		--RaceId	SkillLineId	Race
		[1]  = 60,		--Breton
		[2]  = 62,		--Redguard
		[3]  = 52,		--Orc
		[4]  = 64,		--Dark Elf
		[5]  = 65,		--Nord
		[6]  = 63,		--Argonian
		[7]  = 56,		--High Elf
		[8]  = 57,		--Wood Elf
		[9]  = 58,		--Khajiit
		[10] = 59,		--Imperial
	},
}


local function USPF_RedText(text)	return "|cFF0000"..tostring(text).."|r" end
local function USPF_GreenText(text)	return "|c00FF00"..tostring(text).."|r" end

local function USPF_rgbToHex(rgb)
	local hexStr = '|c'
	for _, v in pairs(rgb) do
		local hex = ''
		local tmpV = math.floor((255 * v) + 0.5)
		while tmpV > 0 do
			local idx = math.fmod(tmpV, 16) + 1
			tmpV = math.floor(tmpV / 16)
			hex = string.sub('0123456789ABCDEF', idx, idx) .. hex
		end
		hex = string.len(hex) == 0 and '00' or (string.len(hex) == 1 and '0' .. hex or hex)
		hexStr = hexStr .. hex
	end
	return hexStr
end

local function USPF_UpdateAllSavedVars()
	if USPF.sVar.ptsData[selectedChar] == nil then USPF.sVar.ptsData[selectedChar] = {} end
	USPF.sVar.ptsData[selectedChar] = USPF_LTF:CopyTable(USPF.ptsData)
end

local function FormatQuestName(questName, completed)
	return completed and "|l0:1:0:-25%:2:ffffff|l"..questName.."|l" or questName
end

local function ColorCompletion(text, completed)
	if completed then
		return USPF_GreenText(text)
	else
		return USPF_RedText(text)
	end
end

local function GetQuestTooltipText(zone)
	local quests = {}
	local isCurrentCharacter = GCCId() == selectedChar
	for _, zoneData in ipairs(USPF.data[zone]) do
		local questName = GetQuestName(zoneData[1])
		local earned = zoneData[2] == 0 and GCQI(zoneData[1]) ~= "" or IAchC(zoneData[2])
		table.insert(quests, FormatQuestName(questName, isCurrentCharacter and earned))
	end
	return table.concat(quests, "\n")
end

local function GetMainQuestTooltip()
	local function FormatProgress(points, total)
		return ColorCompletion(points .."/".. total, points ~= "?" and points >= total)
	end
	local quests = {}
	for _, char in ipairs(USPF.charData) do
		local charPointsData = USPF.sVar.ptsData[char.charId]
		local questPoints = charPointsData.MainQ or "?"
		local questTotal = USPF.ptsTots.MainQ
		table.insert(quests, FormatProgress(questPoints, questTotal) .. "  " .. char.charName)
	end
	return table.concat(quests, "\n")
end

local function GetZoneTooltipText(zone)
    local function FormatProgress(points, total)
        return ColorCompletion(points .."/".. total, points ~= "?" and points >= total)
	end
	local quests = {}
	local zone2 = zone
	if zone == "TG" then zone2 = "HB" end -- different zone names for quests and skyshards in savedVars :-/
	if zone == "RO" then zone2 = "WR" end -- different zone names for quests and skyshards in savedVars :-/
	if zone == "DB" then zone2 = "GC" end -- different zone names for quests and skyshards in savedVars :-/
	if zone == "MW" then zone2 = "VV" end -- different zone names for quests and skyshards in savedVars :-/
	for _, char in ipairs(USPF.charData) do
		local charPointsData = USPF.sVar.ptsData[char.charId]
		local questPoints = charPointsData.ZQ[zone] or "?"
		local questTotal = USPF.ptsTots.ZQ[zone] or 0
		local skyShardPoints = charPointsData.SS[zone2] or "?"
		local skyShardTotal = USPF.ptsTots.SS[zone2]
		local txt = questTotal ~= 0 and FormatProgress(questPoints, questTotal) .. "  " or ""
		table.insert(quests, txt .. FormatProgress(skyShardPoints, skyShardTotal) .. "  " .. char.charName)
	end
	return table.concat(quests, "\n")
end

local function GetGDQuestTooltipText(dungeon)
	local questName = GetQuestName(USPF.data.GD[dungeon])
	local list = {FormatQuestName(questName, selectedChar == GCCId() and GCQI(USPF.data.GD[dungeon]) ~= "").."\n"}
	for _, char in ipairs(USPF.charData) do
		local val = USPF.sVar.ptsData[char.charId].GD[dungeon]
		table.insert(list, ColorCompletion(char.charName, val == 1))
	end
	return table.concat(list, "\n")
end

local function GetAchLink(achId)
	return GetAchievementLink(achId, LINK_TYPE_ACHIEVEMENT)
end

local function GetSV(value)
	return  value ~= nil and value or 0
end

local function GetPDTooltipText(pdung)
	local list = { GetAchLink(USPF.data.PD[pdung]) .. "\n" }
	for _, char in ipairs(USPF.charData) do
		local val = USPF.sVar.ptsData[char.charId].PD[pdung]
		table.insert(list, ColorCompletion(char.charName, val == 1))
	end
	return table.concat(list, "\n")
end

local function getTooltipCharacterTotal()
	local list = {}
	for _, char in ipairs(USPF.charData) do
		local total = USPF.sVar.ptsData[char.charId].Tot
		local unassigned = USPF.sVar.ptsData[char.charId].Unassigned or "?"
		table.insert(list, ColorCompletion(total, total == USPF.ptsTots.Tot) .. "  (" .. unassigned .. ")  " .. char.charName)
	end
	return table.concat(list, "\n")
end

local function getTooltipPDTotal()
	local list = {}
	for _, char in ipairs(USPF.charData) do
		local total = USPF.sVar.ptsData[char.charId].PDTot
		table.insert(list, ColorCompletion(total, total == USPF.ptsTots.PDTot) .. "  " .. char.charName)
	end
	return table.concat(list, "\n")
end

local function getTooltipZoneTotal()
	local list = {}
	for _, char in ipairs(USPF.charData) do
		local questTotal = USPF.sVar.ptsData[char.charId].ZQTot
		local skyShardTotal = USPF.sVar.ptsData[char.charId].SSTot
		table.insert(list, ColorCompletion(questTotal, questTotal == USPF.ptsTots.ZQTot) .. "  " ..
				ColorCompletion(skyShardTotal, skyShardTotal == USPF.ptsTots.SSTot) .. "  " .. char.charName)
	end
	return table.concat(list, "\n")
end

local function getTooltipGDTotal()
	local list = {}
	for _, char in ipairs(USPF.charData) do
		local dungeonTotal = USPF.sVar.ptsData[char.charId].GDTot
		table.insert(list, ColorCompletion(dungeonTotal, dungeonTotal == USPF.ptsTots.GDTot) .. "  " .. char.charName)
	end
	return table.concat(list, "\n")
end


local function getTooltipPvPRank()
	local list = {}
	for _, char in ipairs(USPF.charData) do
		local val = USPF.sVar.ptsData[char.charId].PvPRank
		table.insert(list, ColorCompletion(val, val == USPF.ptsTots.PvPRank) .. "  " .. char.charName)
	end
	return table.concat(list, "\n")
end


local function getTooltipMaelstrom()
	local list = {}
	for _, char in ipairs(USPF.charData) do
		local val = USPF.sVar.ptsData[char.charId].MaelAr
		table.insert(list, ColorCompletion(val, val == 1) .. "  " .. char.charName)
	end
	return table.concat(list, "\n")
end


local function USPF_UpdateGUITable(sVarPtsData)
	local questTooltips = {
		WP	 = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("MQ"),
		AD0	 = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("AD0"),
		AD1  = GetQuestTooltipText("AD1").."\n\n"..GetZoneTooltipText("AD1"),
		AD2  = GetQuestTooltipText("AD2").."\n\n"..GetZoneTooltipText("AD2"),
		AD3  = GetQuestTooltipText("AD3").."\n\n"..GetZoneTooltipText("AD3"),
		AD4  = GetQuestTooltipText("AD4").."\n\n"..GetZoneTooltipText("AD4"),
		AD5  = GetQuestTooltipText("AD5").."\n\n"..GetZoneTooltipText("AD5"),
		CC   = GetQuestTooltipText("CC").."\n\n"..GetZoneTooltipText("CC"),
		CH   = GetQuestTooltipText("CH").."\n\n"..GetZoneTooltipText("CH"),
		CAD	 = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("CAD"),
		CDC	 = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("CDC"),
		CEP	 = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("CEP"),
		CMT	 = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("CMT"),
		DB   = GetQuestTooltipText("DB").."\n\n"..GetZoneTooltipText("DB"),
		DC0a = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("DC0a"),
		DC0b = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("DC0b"),
		DC1	 = GetQuestTooltipText("DC1").."\n\n"..GetZoneTooltipText("DC1"),
		DC2	 = GetQuestTooltipText("DC2").."\n\n"..GetZoneTooltipText("DC2"),
		DC3	 = GetQuestTooltipText("DC3").."\n\n"..GetZoneTooltipText("DC3"),
		DC4	 = GetQuestTooltipText("DC4").."\n\n"..GetZoneTooltipText("DC4"),
		DC5	 = GetQuestTooltipText("DC5").."\n\n"..GetZoneTooltipText("DC5"),
		EO	 = GetQuestTooltipText("EO"),
		EP0a = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("EP0a"),
		EP0b = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("EP0b"),
		EP1	 = GetQuestTooltipText("EP1").."\n\n"..GetZoneTooltipText("EP1"),
		EP2	 = GetQuestTooltipText("EP1").."\n\n"..GetZoneTooltipText("EP2"),
		EP3	 = GetQuestTooltipText("EP2").."\n\n"..GetZoneTooltipText("EP3"),
		EP4	 = GetQuestTooltipText("EP3").."\n\n"..GetZoneTooltipText("EP4"),
		EP5	 = GetQuestTooltipText("EP4").."\n\n"..GetZoneTooltipText("EP5"),
		IC	 = GetQuestTooltipText("IC").."\n\n"..GetZoneTooltipText("IC"),
		LCL	 = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("LCL"),
		UCL	 = GS(USPF_QUEST_NONE).."\n\n"..GetZoneTooltipText("UCL"),
		MM	 = GetQuestTooltipText("MM").."\n\n"..GetZoneTooltipText("MM"),
		MO	 = GetQuestTooltipText("MO"),
		MW	 = GetQuestTooltipText("MW").."\n\n"..GetZoneTooltipText("MW"),
		MQ	 = GetQuestTooltipText("MQ").."\n\n"..GetMainQuestTooltip(),
		NE	 = GetQuestTooltipText("NE").."\n\n"..GetZoneTooltipText("NE"),
		RO	 = GetQuestTooltipText("RO").."\n\n"..GetZoneTooltipText("RO"),
		SE	 = GetQuestTooltipText("SE").."\n\n"..GetZoneTooltipText("SE"),
		SO	 = GetQuestTooltipText("SO"),
		TG	 = GetQuestTooltipText("TG").."\n\n"..GetZoneTooltipText("TG"),
		SU	 = GetQuestTooltipText("SU").."\n\n"..GetZoneTooltipText("SU"),
		WS   = GetQuestTooltipText("WS").."\n\n"..GetZoneTooltipText("WS"),
		GO   = GetQuestTooltipText("GO"),
		BO   = GetQuestTooltipText("BO"),
		TR   = GetQuestTooltipText("TR").."\n\n"..GetZoneTooltipText("TR"),
		BW   = GetQuestTooltipText("BW").."\n\n"..GetZoneTooltipText("BW"),
	}

	USPF.GUI = {
		GSP = {
			{ 1, GS(USPF_GUI_CHAR_LEVEL),	GetSV(sVarPtsData.Level),	USPF.ptsTots.Level,		GS(USPF_QUEST_NA)},
			{ 2, GS(USPF_GUI_MAIN_QUEST),	GetSV(sVarPtsData.MainQ),	USPF.ptsTots.MainQ,		questTooltips.MQ},
			{ 3, GS(USPF_GUI_FOLIUM),		GetSV(sVarPtsData.FolDis),	USPF.ptsTots.FolDis,	GS(USPF_QUEST_NA)},
			{ 4, GS(USPF_GUI_MW_CHAR),		GetSV(sVarPtsData.MWChar),	USPF.ptsTots.MWChar,	questTooltips.MO},
			{ 5, GS(USPF_GUI_SU_CHAR),		GetSV(sVarPtsData.SUChar),	USPF.ptsTots.SUChar,	questTooltips.SO},
			{ 6, GS(USPF_GUI_EW_CHAR),		GetSV(sVarPtsData.EWChar),	USPF.ptsTots.EWChar,	questTooltips.EO},
			{ 7, GS(USPF_GUI_GM_CHAR),		GetSV(sVarPtsData.GMChar),	USPF.ptsTots.GMChar,	questTooltips.GO},
			{ 8, GS(USPF_GUI_BW_CHAR),		GetSV(sVarPtsData.BWChar),	USPF.ptsTots.BWChar,	questTooltips.BO},
			{ 9, GS(USPF_GUI_AVA_RANK),		GetSV(sVarPtsData.PvPRank),	USPF.ptsTots.PvPRank,	getTooltipPvPRank()},
			{10, GS(USPF_GUI_MAEL_ARENA),	GetSV(sVarPtsData.MaelAr),	USPF.ptsTots.MaelAr,	getTooltipMaelstrom()},
		},
		GSP_T = strF("%s: %d/%d", GS(USPF_GUI_TOTAL), sVarPtsData.GenTot, USPF.ptsTots.GenTot),
		SQS = {
			{ 1, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.WP)),						GetSV(sVarPtsData.ZQ.WP),	USPF.ptsTots.ZQ.WP,		GetSV(sVarPtsData.SS.MQ),	USPF.ptsTots.SS.MQ,		questTooltips.WP},
			{ 2, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.AD0)),						GetSV(sVarPtsData.ZQ.AD0),	USPF.ptsTots.ZQ.AD0,	GetSV(sVarPtsData.SS.AD0),	USPF.ptsTots.SS.AD0,	questTooltips.AD0},
			{ 3, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.AD1)),						GetSV(sVarPtsData.ZQ.AD1),	USPF.ptsTots.ZQ.AD1,	GetSV(sVarPtsData.SS.AD1),	USPF.ptsTots.SS.AD1,	questTooltips.AD1},
			{ 4, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.AD2)),						GetSV(sVarPtsData.ZQ.AD2),	USPF.ptsTots.ZQ.AD2,	GetSV(sVarPtsData.SS.AD2),	USPF.ptsTots.SS.AD2,	questTooltips.AD2},
			{ 5, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.AD3)),						GetSV(sVarPtsData.ZQ.AD3),	USPF.ptsTots.ZQ.AD3,	GetSV(sVarPtsData.SS.AD3),	USPF.ptsTots.SS.AD3,	questTooltips.AD3},
			{ 6, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.AD4)),						GetSV(sVarPtsData.ZQ.AD4),	USPF.ptsTots.ZQ.AD4,	GetSV(sVarPtsData.SS.AD4),	USPF.ptsTots.SS.AD4,	questTooltips.AD4},
			{ 7, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.AD5)),						GetSV(sVarPtsData.ZQ.AD5),	USPF.ptsTots.ZQ.AD5,	GetSV(sVarPtsData.SS.AD5),	USPF.ptsTots.SS.AD5,	questTooltips.AD5},
			{ 8, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.DC0B)),						GetSV(sVarPtsData.ZQ.DC0b),	USPF.ptsTots.ZQ.DC0b,	GetSV(sVarPtsData.SS.DC0b),	USPF.ptsTots.SS.DC0b,	questTooltips.DC0b},
			{ 9, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.DC0A)),						GetSV(sVarPtsData.ZQ.DC0a),	USPF.ptsTots.ZQ.DC0a,	GetSV(sVarPtsData.SS.DC0a),	USPF.ptsTots.SS.DC0a,	questTooltips.DC0a},
			{10, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.DC1)),						GetSV(sVarPtsData.ZQ.DC1),	USPF.ptsTots.ZQ.DC1,	GetSV(sVarPtsData.SS.DC1),	USPF.ptsTots.SS.DC1,	questTooltips.DC1},
			{11, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.DC2)),						GetSV(sVarPtsData.ZQ.DC2),	USPF.ptsTots.ZQ.DC2,	GetSV(sVarPtsData.SS.DC2),	USPF.ptsTots.SS.DC2,	questTooltips.DC2},
			{12, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.DC3)),						GetSV(sVarPtsData.ZQ.DC3),	USPF.ptsTots.ZQ.DC3,	GetSV(sVarPtsData.SS.DC3),	USPF.ptsTots.SS.DC3,	questTooltips.DC3},
			{13, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.DC4)),						GetSV(sVarPtsData.ZQ.DC4),	USPF.ptsTots.ZQ.DC4,	GetSV(sVarPtsData.SS.DC4),	USPF.ptsTots.SS.DC4,	questTooltips.DC4},
			{14, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.DC5)),						GetSV(sVarPtsData.ZQ.DC5),	USPF.ptsTots.ZQ.DC5,	GetSV(sVarPtsData.SS.DC5),	USPF.ptsTots.SS.DC5,	questTooltips.DC5},
			{15, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.EP0B)),						GetSV(sVarPtsData.ZQ.EP0b),	USPF.ptsTots.ZQ.EP0b,	GetSV(sVarPtsData.SS.EP0b),	USPF.ptsTots.SS.EP0b,	questTooltips.EP0b},
			{16, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.EP0A)),						GetSV(sVarPtsData.ZQ.EP0a),	USPF.ptsTots.ZQ.EP0a,	GetSV(sVarPtsData.SS.EP0a),	USPF.ptsTots.SS.EP0a,	questTooltips.EP0a},
			{17, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.EP1)),						GetSV(sVarPtsData.ZQ.EP1),	USPF.ptsTots.ZQ.EP1,	GetSV(sVarPtsData.SS.EP1),	USPF.ptsTots.SS.EP1,	questTooltips.EP1},
			{18, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.EP2)),						GetSV(sVarPtsData.ZQ.EP2),	USPF.ptsTots.ZQ.EP2,	GetSV(sVarPtsData.SS.EP2),	USPF.ptsTots.SS.EP2,	questTooltips.EP2},
			{19, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.EP3)),						GetSV(sVarPtsData.ZQ.EP3),	USPF.ptsTots.ZQ.EP3,	GetSV(sVarPtsData.SS.EP3),	USPF.ptsTots.SS.EP3,	questTooltips.EP3},
			{20, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.EP4)),						GetSV(sVarPtsData.ZQ.EP4),	USPF.ptsTots.ZQ.EP4,	GetSV(sVarPtsData.SS.EP4),	USPF.ptsTots.SS.EP4,	questTooltips.EP4},
			{21, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.EP5)),						GetSV(sVarPtsData.ZQ.EP5),	USPF.ptsTots.ZQ.EP5,	GetSV(sVarPtsData.SS.EP5),	USPF.ptsTots.SS.EP5,	questTooltips.EP5},
			{22, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.CH)),						GetSV(sVarPtsData.ZQ.CH),	USPF.ptsTots.ZQ.CH,		GetSV(sVarPtsData.SS.CH),	USPF.ptsTots.SS.CH,		questTooltips.CH},
			{23, zf("<<C:1>> <<2>>",	GZNBId(USPF.data.ZId.ZN.CAD), "AD"),				GetSV(sVarPtsData.ZQ.CAD),	USPF.ptsTots.ZQ.CAD,	GetSV(sVarPtsData.SS.CAD),	USPF.ptsTots.SS.CAD,	questTooltips.CAD},
			{24, zf("<<C:1>> <<2>>",	GZNBId(USPF.data.ZId.ZN.CDC), "DC"),				GetSV(sVarPtsData.ZQ.CDC),	USPF.ptsTots.ZQ.CDC,	GetSV(sVarPtsData.SS.CDC),	USPF.ptsTots.SS.CDC,	questTooltips.CDC},
			{25, zf("<<C:1>> <<2>>",	GZNBId(USPF.data.ZId.ZN.CEP), "EP"),				GetSV(sVarPtsData.ZQ.CEP),	USPF.ptsTots.ZQ.CEP,	GetSV(sVarPtsData.SS.CEP),	USPF.ptsTots.SS.CEP,	questTooltips.CEP},
			{26, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.CMT)),						GetSV(sVarPtsData.ZQ.CMT),	USPF.ptsTots.ZQ.CMT,	GetSV(sVarPtsData.SS.CMT),	USPF.ptsTots.SS.CMT,	questTooltips.CMT},
			{27, zf("<<C:1>> <<C:2>>",	GS(USPF_GUI_ZN_LCL), GZNBId(USPF.data.ZId.ZN.CL)),	GetSV(sVarPtsData.ZQ.LCL),	USPF.ptsTots.ZQ.LCL,	GetSV(sVarPtsData.SS.LCL),	USPF.ptsTots.SS.LCL,	questTooltips.LCL},
			{28, zf("<<C:1>> <<C:2>>",	GS(USPF_GUI_ZN_UCL), GZNBId(USPF.data.ZId.ZN.CL)),	GetSV(sVarPtsData.ZQ.UCL),	USPF.ptsTots.ZQ.UCL,	GetSV(sVarPtsData.SS.UCL),	USPF.ptsTots.SS.UCL,	questTooltips.UCL},
			{29, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.IC)),						GetSV(sVarPtsData.ZQ.IC),	USPF.ptsTots.ZQ.IC,		GetSV(sVarPtsData.SS.IC),	USPF.ptsTots.SS.IC,		questTooltips.IC},
			{30, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.RO)),						GetSV(sVarPtsData.ZQ.RO),	USPF.ptsTots.ZQ.RO,		GetSV(sVarPtsData.SS.WR),	USPF.ptsTots.SS.WR,		questTooltips.RO},
			{31, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.HB)),						GetSV(sVarPtsData.ZQ.TG),	USPF.ptsTots.ZQ.TG,		GetSV(sVarPtsData.SS.HB),	USPF.ptsTots.SS.HB,		questTooltips.TG},
			{32, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.GC)),						GetSV(sVarPtsData.ZQ.DB),	USPF.ptsTots.ZQ.DB,		GetSV(sVarPtsData.SS.GC),	USPF.ptsTots.SS.GC,		questTooltips.DB},
			{33, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.VV)),						GetSV(sVarPtsData.ZQ.MW),	USPF.ptsTots.ZQ.MW,		GetSV(sVarPtsData.SS.VV),	USPF.ptsTots.SS.VV,		questTooltips.MW},
			{34, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.CC)),						GetSV(sVarPtsData.ZQ.CC),	USPF.ptsTots.ZQ.CC,		GetSV(sVarPtsData.SS.CC),	USPF.ptsTots.SS.CC,		questTooltips.CC},
			{35, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.SU)),						GetSV(sVarPtsData.ZQ.SU),	USPF.ptsTots.ZQ.SU,		GetSV(sVarPtsData.SS.SU),	USPF.ptsTots.SS.SU,		questTooltips.SU},
			{36, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.MM)),						GetSV(sVarPtsData.ZQ.MM),	USPF.ptsTots.ZQ.MM,		GetSV(sVarPtsData.SS.MM),	USPF.ptsTots.SS.MM,		questTooltips.MM},
			{37, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.NE)),						GetSV(sVarPtsData.ZQ.NE),	USPF.ptsTots.ZQ.NE,		GetSV(sVarPtsData.SS.NE),	USPF.ptsTots.SS.NE,		questTooltips.NE},
			{38, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.SE)),						GetSV(sVarPtsData.ZQ.SE),	USPF.ptsTots.ZQ.SE,		GetSV(sVarPtsData.SS.SE),	USPF.ptsTots.SS.SE,		questTooltips.SE},
			{39, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.WS)),						GetSV(sVarPtsData.ZQ.WS),	USPF.ptsTots.ZQ.WS,		GetSV(sVarPtsData.SS.WS),	USPF.ptsTots.SS.WS,		questTooltips.WS},
			{40, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.TR)),						GetSV(sVarPtsData.ZQ.TR),	USPF.ptsTots.ZQ.TR,		GetSV(sVarPtsData.SS.TR),	USPF.ptsTots.SS.TR,		questTooltips.TR},
			{41, zf("<<C:1>>",			GZNBId(USPF.data.ZId.ZN.BW)),						GetSV(sVarPtsData.ZQ.BW),	USPF.ptsTots.ZQ.BW,		GetSV(sVarPtsData.SS.BW),	USPF.ptsTots.SS.BW,		questTooltips.BW},
		},
		SQS_SL_T = strF("%d/%d", sVarPtsData.ZQTot, USPF.ptsTots.ZQTot),
		SQS_SS_T = strF("%d/%d", sVarPtsData.SSTot, USPF.ptsTots.SSTot),
		GDQ = {
			{ 1, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.BC1)),	GetSV(sVarPtsData.GD.BC1),	USPF.ptsTots.GD.BC1,	GetGDQuestTooltipText("BC1")},
			{ 2, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.BC2)),	GetSV(sVarPtsData.GD.BC2),	USPF.ptsTots.GD.BC2,	GetGDQuestTooltipText("BC2")},
			{ 3, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.EH1)),	GetSV(sVarPtsData.GD.EH1),	USPF.ptsTots.GD.EH1,	GetGDQuestTooltipText("EH1")},
			{ 4, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.EH2)),	GetSV(sVarPtsData.GD.EH2),	USPF.ptsTots.GD.EH2,	GetGDQuestTooltipText("EH2")},
			{ 5, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.CA1)),	GetSV(sVarPtsData.GD.CA1),	USPF.ptsTots.GD.CA1,	GetGDQuestTooltipText("CA1")},
			{ 6, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.CA2)),	GetSV(sVarPtsData.GD.CA2),	USPF.ptsTots.GD.CA2,	GetGDQuestTooltipText("CA2")},
			{ 7, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD4)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.TI)),	GetSV(sVarPtsData.GD.TI),	USPF.ptsTots.GD.TI,		GetGDQuestTooltipText("TI")},
			{ 8, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.SW)),	GetSV(sVarPtsData.GD.SW),	USPF.ptsTots.GD.SW,		GetGDQuestTooltipText("SW")},
			{ 9, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.SC1)),	GetSV(sVarPtsData.GD.SC1),	USPF.ptsTots.GD.SC1,	GetGDQuestTooltipText("SC1")},
			{10, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.SC2)),	GetSV(sVarPtsData.GD.SC2),	USPF.ptsTots.GD.SC2,	GetGDQuestTooltipText("SC2")},
			{11, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.WS1)),	GetSV(sVarPtsData.GD.WS1),	USPF.ptsTots.GD.WS1,	GetGDQuestTooltipText("WS1")},
			{12, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.WS2)),	GetSV(sVarPtsData.GD.WS2),	USPF.ptsTots.GD.WS2,	GetGDQuestTooltipText("WS2")},
			{13, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.CH1)),	GetSV(sVarPtsData.GD.CH1),	USPF.ptsTots.GD.CH1,	GetGDQuestTooltipText("CH1")},
			{14, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.CH2)),	GetSV(sVarPtsData.GD.CH2),	USPF.ptsTots.GD.CH2,	GetGDQuestTooltipText("CH2")},
			{15, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC4)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.VF)),	GetSV(sVarPtsData.GD.VF),	USPF.ptsTots.GD.VF,		GetGDQuestTooltipText("VF")},
			{16, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.BH)),	GetSV(sVarPtsData.GD.BH),	USPF.ptsTots.GD.BH,		GetGDQuestTooltipText("BH")},
			{17, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.FG1)),	GetSV(sVarPtsData.GD.FG1),	USPF.ptsTots.GD.FG1,	GetGDQuestTooltipText("FG1")},
			{18, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.FG2)),	GetSV(sVarPtsData.GD.FG2),	USPF.ptsTots.GD.FG2,	GetGDQuestTooltipText("FG2")},
			{19, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.DC1)),	GetSV(sVarPtsData.GD.DC1),	USPF.ptsTots.GD.DC1,	GetGDQuestTooltipText("DC1")},
			{20, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.DC2)),	GetSV(sVarPtsData.GD.DC2),	USPF.ptsTots.GD.DC2,	GetGDQuestTooltipText("DC2")},
			{21, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.AC)),	GetSV(sVarPtsData.GD.AC),	USPF.ptsTots.GD.AC,		GetGDQuestTooltipText("AC")},
			{22, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP4)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.DK)),	GetSV(sVarPtsData.GD.DK),	USPF.ptsTots.GD.DK,		GetGDQuestTooltipText("DK")},
			{23, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.BC)),	GetSV(sVarPtsData.GD.BC),	USPF.ptsTots.GD.BC,		GetGDQuestTooltipText("BC")},
			{24, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.CH)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.VM)),	GetSV(sVarPtsData.GD.VM),	USPF.ptsTots.GD.VM,		GetGDQuestTooltipText("VM")},
			{25, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.CYD)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.ICP)),	GetSV(sVarPtsData.GD.ICP),	USPF.ptsTots.GD.ICP,	GetGDQuestTooltipText("ICP")},
			{26, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.CYD)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.WGT)),	GetSV(sVarPtsData.GD.WGT),	USPF.ptsTots.GD.WGT,	GetGDQuestTooltipText("WGT")},
			{27, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.CS)),	GetSV(sVarPtsData.GD.CS),	USPF.ptsTots.GD.CS,		GetGDQuestTooltipText("CS")},
			{28, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.RM)),	GetSV(sVarPtsData.GD.RM),	USPF.ptsTots.GD.RM,		GetGDQuestTooltipText("RM")},
			{29, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.CL)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.BF)),	GetSV(sVarPtsData.GD.BF),	USPF.ptsTots.GD.BF,		GetGDQuestTooltipText("BF")},
			{30, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.CL)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.FH)),	GetSV(sVarPtsData.GD.FH),	USPF.ptsTots.GD.FH,		GetGDQuestTooltipText("FH")},
			{31, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.FL)),	GetSV(sVarPtsData.GD.FL),	USPF.ptsTots.GD.FL,		GetGDQuestTooltipText("FL")},
			{32, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.SP)),	GetSV(sVarPtsData.GD.SP),	USPF.ptsTots.GD.SP,		GetGDQuestTooltipText("SP")},
			{33, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.MHK)),	GetSV(sVarPtsData.GD.MHK),	USPF.ptsTots.GD.MHK,	GetGDQuestTooltipText("MHK")},
			{34, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.MOS)),	GetSV(sVarPtsData.GD.MOS),	USPF.ptsTots.GD.MOS,	GetGDQuestTooltipText("MOS")},
			{35, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.GC)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.DoM)),	GetSV(sVarPtsData.GD.DoM),	USPF.ptsTots.GD.DoM,	GetGDQuestTooltipText("DoM")},
			{36, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP4)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.FV)),	GetSV(sVarPtsData.GD.FV),	USPF.ptsTots.GD.FV,		GetGDQuestTooltipText("FV")},
			{37, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.LM)),	GetSV(sVarPtsData.GD.LM),	USPF.ptsTots.GD.LM,		GetGDQuestTooltipText("LM")},
			{38, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.NE)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.MF)),	GetSV(sVarPtsData.GD.MF),	USPF.ptsTots.GD.MF,		GetGDQuestTooltipText("MF")},
			{39, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.RO)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.IR)),	GetSV(sVarPtsData.GD.IR),	USPF.ptsTots.GD.IR,		GetGDQuestTooltipText("IR")},
			{40, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.UG)),	GetSV(sVarPtsData.GD.UG),	USPF.ptsTots.GD.UG,		GetGDQuestTooltipText("UG")},
			{41, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.BGC)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.SG)),	GetSV(sVarPtsData.GD.SG),	USPF.ptsTots.GD.SG,		GetGDQuestTooltipText("SG")},
			{42, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.WS)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.CT)),	GetSV(sVarPtsData.GD.CT),	USPF.ptsTots.GD.CT,		GetGDQuestTooltipText("CT")},
			{43, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.GC)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.BDV)),	GetSV(sVarPtsData.GD.BDV),	USPF.ptsTots.GD.BDV,	GetGDQuestTooltipText("BDV")},
			{44, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.TC)),	GetSV(sVarPtsData.GD.TC),	USPF.ptsTots.GD.TC,		GetGDQuestTooltipText("TC")},
			{45, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.RPB)),	GetSV(sVarPtsData.GD.RPB),	USPF.ptsTots.GD.RPB,	GetGDQuestTooltipText("RPB")},
			{46, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.BW)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.GDN.TDC)),	GetSV(sVarPtsData.GD.TDC),	USPF.ptsTots.GD.TDC,	GetGDQuestTooltipText("TDC")},
		},
		GDQ_T = strF("%s: %d/%d", GS(USPF_GUI_TOTAL), sVarPtsData.GDTot, USPF.ptsTots.GDTot),
		PDGBE = {
			{ 1, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.AD1)),	GetSV(sVarPtsData.PD.AD1),	USPF.ptsTots.PD.AD1,	GetPDTooltipText("AD1")},
			{ 2, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.AD2)),	GetSV(sVarPtsData.PD.AD2),	USPF.ptsTots.PD.AD2,	GetPDTooltipText("AD2")},
			{ 3, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.AD3)),	GetSV(sVarPtsData.PD.AD3),	USPF.ptsTots.PD.AD3,	GetPDTooltipText("AD3")},
			{ 4, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD4)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.AD4)),	GetSV(sVarPtsData.PD.AD4),	USPF.ptsTots.PD.AD4,	GetPDTooltipText("AD4")},
			{ 5, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.AD5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.AD5)),	GetSV(sVarPtsData.PD.AD5),	USPF.ptsTots.PD.AD5,	GetPDTooltipText("AD5")},
			{ 6, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.DC1)),	GetSV(sVarPtsData.PD.DC1),	USPF.ptsTots.PD.DC1,	GetPDTooltipText("DC1")},
			{ 7, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.DC2)),	GetSV(sVarPtsData.PD.DC2),	USPF.ptsTots.PD.DC2,	GetPDTooltipText("DC2")},
			{ 8, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.DC3)),	GetSV(sVarPtsData.PD.DC3),	USPF.ptsTots.PD.DC3,	GetPDTooltipText("DC3")},
			{ 9, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC4)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.DC4)),	GetSV(sVarPtsData.PD.DC4),	USPF.ptsTots.PD.DC4,	GetPDTooltipText("DC4")},
			{10, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.DC5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.DC5)),	GetSV(sVarPtsData.PD.DC5),	USPF.ptsTots.PD.DC5,	GetPDTooltipText("DC5")},
			{11, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP1)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.EP1)),	GetSV(sVarPtsData.PD.EP1),	USPF.ptsTots.PD.EP1,	GetPDTooltipText("EP1")},
			{12, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP2)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.EP2)),	GetSV(sVarPtsData.PD.EP2),	USPF.ptsTots.PD.EP2,	GetPDTooltipText("EP2")},
			{13, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP3)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.EP3)),	GetSV(sVarPtsData.PD.EP3),	USPF.ptsTots.PD.EP3,	GetPDTooltipText("EP3")},
			{14, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP4)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.EP4)),	GetSV(sVarPtsData.PD.EP4),	USPF.ptsTots.PD.EP4,	GetPDTooltipText("EP4")},
			{15, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.EP5)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.EP5)),	GetSV(sVarPtsData.PD.EP5),	USPF.ptsTots.PD.EP5,	GetPDTooltipText("EP5")},
			{16, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.CH)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.CH)),	GetSV(sVarPtsData.PD.CH),		USPF.ptsTots.PD.CH,		GetPDTooltipText("CH")},
			{17, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.RO)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.WOO)),	GetSV(sVarPtsData.PD.WOO),	USPF.ptsTots.PD.WOO,	GetPDTooltipText("WOO")},
			{18, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.RO)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.WRK)),	GetSV(sVarPtsData.PD.WRK),	USPF.ptsTots.PD.WRK,	GetPDTooltipText("WRK")},
			{19, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.VV)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.VFW)),	GetSV(sVarPtsData.PD.VFW),	USPF.ptsTots.PD.VFW,	GetPDTooltipText("VFW")},
			{20, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.VV)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.VNC)),	GetSV(sVarPtsData.PD.VNC),	USPF.ptsTots.PD.VNC,	GetPDTooltipText("VNC")},
			{21, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.SU)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.SKW)),	GetSV(sVarPtsData.PD.SKW),	USPF.ptsTots.PD.SKW,	GetPDTooltipText("SKW")},
			{22, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.SU)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.SSH)),	GetSV(sVarPtsData.PD.SSH),	USPF.ptsTots.PD.SSH,	GetPDTooltipText("SSH")},
			{23, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.NE)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.OC)),	GetSV(sVarPtsData.PD.OC),		USPF.ptsTots.PD.OC,		GetPDTooltipText("OC")},
			{24, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.NE)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.RN)),	GetSV(sVarPtsData.PD.RN),		USPF.ptsTots.PD.RN,		GetPDTooltipText("RN")},
			{25, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.WS)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.LT)),	GetSV(sVarPtsData.PD.LT),		USPF.ptsTots.PD.LT,		GetPDTooltipText("LT")},
			{26, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.BGC)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.NK)),	GetSV(sVarPtsData.PD.NK),		USPF.ptsTots.PD.NK,		GetPDTooltipText("NK")},
			{27, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.BW)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.SH)),	GetSV(sVarPtsData.PD.SH),		USPF.ptsTots.PD.SH,		GetPDTooltipText("SH")},
			{28, zf("<<C:1>>", GZNBId(USPF.data.ZId.ZN.BW)),	zf("<<C:1>>", GZNBId(USPF.data.ZId.PDN.ZA)),	GetSV(sVarPtsData.PD.ZA),		USPF.ptsTots.PD.ZA,		GetPDTooltipText("ZA")},
		},
		PDGBE_T = strF("%s: %d/%d", GS(USPF_GUI_TOTAL), sVarPtsData.PDTot, USPF.ptsTots.PDTot),
		CharacterTot = strF("%s: %d/%d (%s unassigned)", GS(USPF_GUI_CHAR_TOTAL), sVarPtsData.Tot, USPF.ptsTots.Tot, sVarPtsData.Unassigned and tostring(sVarPtsData.Unassigned) or "?"),
	}
end

local function USPF_CheckSavedVars(value)
	local charId = GCCId()
	if selectedChar ~= charId then return false end
	if value == nil then
		--Write the character settings table.
		USPF.sVar.settings[charId] = {}
		USPF.sVar.settings[charId] = USPF_LTF:CopyTable(USPF.settings)

		--Write the character points data table.
		USPF.sVar.ptsData[charId] = {}
		USPF.sVar.ptsData[charId] = USPF_LTF:CopyTable(USPF.ptsData)
	end
	return true
end

local function USPF_SetLevelPoints()
	local level = GetUnitLevel("player")
	USPF.ptsData.Level = math.floor(level/5) + math.floor(level/10) + (level - 1)

	--Update saved variables.
	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		USPF.sVar.ptsData[selectedChar].Level = USPF.ptsData.Level
	end
end

local function USPF_SetQuestPoints()
	--Main Quest Skill Points
	for i=1, #USPF.data.MQ do
		USPF.ptsData.MainQ = USPF.ptsData.MainQ + (IAchC(USPF.data.MQ[i][2]) and 1 or 0)
	end

	--Morrowind Only Character Quest Skill Points
	USPF.ptsData.MWChar = (GCQI(USPF.data.MO[1][1]) ~= "" or USPF.settings.MWC) and 1 or 0

	--Summerset Only Character Quest Skill Points
	USPF.ptsData.SUChar = (GCQI(USPF.data.SO[1][1]) ~= "" or USPF.settings.SSC) and 1 or 0

	--Elsweyr Only Character Quest Skill Points
	USPF.ptsData.EWChar = (GCQI(USPF.data.EO[1][1]) ~= "" or USPF.settings.EWC) and 1 or 0

	--Greymoor Only Character Quest Skill Points
	USPF.ptsData.GMChar = (GCQI(USPF.data.GO[1][1]) ~= "" or USPF.settings.GMC) and 1 or 0

	--Blackwood Only Character Quest Skill Points
	USPF.ptsData.BWChar = (GCQI(USPF.data.BO[1][1]) ~= "" or USPF.settings.BWC) and 1 or 0

	for k,_ in pairs(USPF.ptsData.ZQ) do
		if USPF.data[k] ~= nil then
			for i=1, #USPF.data[k] do
				USPF.ptsData.ZQ[k] = USPF.ptsData.ZQ[k] + ((USPF.data[k][i][2] == 0 and (GCQI(USPF.data[k][i][1]) ~= "") or IAchC(USPF.data[k][i][2])) and 1 or 0)
			end
			USPF.ptsData.ZQTot = USPF.ptsData.ZQTot + USPF.ptsData.ZQ[k]
		end
	end

	--Group Dungeon Quest Skill Points
	for k,_ in pairs(USPF.ptsData.GD) do
		USPF.ptsData.GD[k] = GCQI(USPF.data.GD[k]) ~= "" and 1 or 0
		USPF.ptsData.GDTot = USPF.ptsData.GDTot + USPF.ptsData.GD[k]
	end

	--Update saved variables for all.
	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		USPF.sVar.ptsData[selectedChar].MainQ = USPF.ptsData.MainQ
		USPF.sVar.ptsData[selectedChar].MWChar = USPF.ptsData.MWChar
		USPF.sVar.ptsData[selectedChar].SUChar = USPF.ptsData.SUChar
		USPF.sVar.ptsData[selectedChar].EWChar = USPF.ptsData.EWChar
		USPF.sVar.ptsData[selectedChar].GMChar = USPF.ptsData.GMChar
		USPF.sVar.ptsData[selectedChar].BWChar = USPF.ptsData.BWChar

		USPF.sVar.ptsData[selectedChar].ZQ = USPF_LTF:CopyTable(USPF.ptsData.ZQ)
		USPF.sVar.ptsData[selectedChar].GD = USPF_LTF:CopyTable(USPF.ptsData.GD)

		USPF.sVar.ptsData[selectedChar].ZQTot = USPF.ptsData.ZQTot
		USPF.sVar.ptsData[selectedChar].GDTot = USPF.ptsData.GDTot
	end
end

local function USPF_SetPublicDungeonPoints()
	for k,_ in pairs(USPF.ptsData.PD) do
		USPF.ptsData.PD[k] = IAchC(USPF.data.PD[k]) and 1 or 0
		USPF.ptsData.PDTot = USPF.ptsData.PDTot + USPF.ptsData.PD[k]
	end

	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		--Update saved variables.
		USPF.sVar.ptsData[selectedChar].PD		= USPF_LTF:CopyTable(USPF.ptsData.PD)
		USPF.sVar.ptsData[selectedChar].PDTot	= USPF.ptsData.PDTot
	end
end

local function USPF_SetSkyshardPoints()
	for k,_ in pairs(USPF.ptsData.SS) do
		if IAchC(USPF.data.SS[k]) then USPF.ptsData.SS[k] = GAchNCr(USPF.data.SS[k])
		else
			for i = 1, GAchNCr(USPF.data.SS[k]) do
				local _, numCompleted = GAchCr(USPF.data.SS[k], i)
				if numCompleted > 0 then USPF.ptsData.SS[k] = USPF.ptsData.SS[k] + 1 end
			end
		end
		USPF.ptsData.numSSTot = USPF.ptsData.numSSTot + USPF.ptsData.SS[k]
	end

	--Calculate the total and round for points.
	USPF.ptsData.SSTot = math.floor(USPF.ptsData.numSSTot/3)

	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		--Update saved variables for all.
		USPF.sVar.ptsData[selectedChar].SS			= USPF_LTF:CopyTable(USPF.ptsData.SS)
		USPF.sVar.ptsData[selectedChar].numSSTot	= USPF.ptsData.numSSTot
		USPF.sVar.ptsData[selectedChar].SSTot		= USPF.ptsData.SSTot
	end
end

local function USPF_IsValidRacialLine(skillType, skillLine)
	if skillType == SKILL_TYPE_RACIAL then
		local _, _, _, skillLineId = GetSkillLineInfo(skillType, skillLine)
		return skillLineId == USPF.data.racialLineIds[GetUnitRaceId("player")]
	end
	return false
end

local function USPF_GetSkillSpentPoints(skillType, skillLine, skillIndex)
	local skills = USPF.cache.skillTypes[skillType].lines[skillLine].skills
	local _, _, _, _, _, purchased, progressionIndex = GetSkillAbilityInfo(skillType, skillLine, skillIndex)
	local spent, possible, reduction

	if not purchased then spent = 0 end
	reduction = IsSkillAbilityAutoGrant(skillType, skillLine, skillIndex) and 1 or 0

	if progressionIndex then
		local _, morph = GetAbilityProgressionInfo(progressionIndex)
		if not spent then
			spent = morph > 0 and (2 - reduction) or (1 - reduction)
		end
		possible = 2 - reduction
	else
		local currentLevel, maxLevel = GetSkillAbilityUpgradeInfo(skillType, skillLine, skillIndex)
		if currentLevel then
			if not spent then spent = currentLevel - reduction end
			possible = maxLevel - reduction
		else
			if not spent then spent = 1 - reduction end
			possible = 1 - reduction
		end
	end
	skills[skillIndex] = {spent, possible}
	return skills[skillIndex]
end

local function USPF_GetLineSpentPoints(skillType, skillLine)
	if skillType ~= SKILL_TYPE_RACIAL or USPF_IsValidRacialLine(skillType, skillLine) then
		USPF.cache.skillTypes[skillType].lines[skillLine] = {}
		local line = USPF.cache.skillTypes[skillType].lines[skillLine]
		line.skills = {}
		line.total = 0
		line.possible = 0
		local numSkillAbilities = GetNumSkillAbilities(skillType, skillLine)
		for ability = 1, numSkillAbilities do
			local spent = USPF_GetSkillSpentPoints(skillType, skillLine, ability)
			line.total = line.total + spent[1]
			line.possible = line.possible + spent[2]
		end

		return line.total
	end
	return 0
end

local function USPF_GetTypeSpentPoints(skillTypeId)
	USPF.cache.skillTypes[skillTypeId] = {}
	local skillType = USPF.cache.skillTypes[skillTypeId]
	skillType.lines = {}
	skillType.total = 0
	local numSkillLines = GetNumSkillLines(skillTypeId)
	for skillLine = 1, numSkillLines do skillType.total = skillType.total + USPF_GetLineSpentPoints(skillTypeId, skillLine) end
	return skillType.total
end

local function USPF_GetTotSkillPoints()
	USPF.cache.skillTypes = {}
	USPF.cache.total = GetAvailableSkillPoints()

	for skillType = 1, GetNumSkillTypes() do USPF.cache.total = USPF.cache.total + USPF_GetTypeSpentPoints(skillType) end
	return USPF.cache.total
end

local function USPF_SetFoliumDiscognitumPoints()
	if USPF.settings.FD.override then
		USPF.ptsData.FolDis = USPF.settings.FD.charHasFD and 2 or 0
	else
		local skillPoints = USPF_GetTotSkillPoints()
		local skillPointsDiff =	USPF.ptsData.Level + USPF.ptsData.MainQ + USPF.ptsData.MWChar +
								USPF.ptsData.SUChar + USPF.ptsData.EWChar + USPF.ptsData.GMChar +
								USPF.ptsData.BWChar + USPF.ptsData.PvPRank + USPF.ptsData.MaelAr +
								USPF.ptsData.ZQTot + USPF.ptsData.SSTot + USPF.ptsData.GDTot +
								USPF.ptsData.PDTot
		USPF.ptsData.FolDis = skillPoints == skillPointsDiff + 2 and 2 or 0
	end

	--Update saved variables.
	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		USPF.sVar.ptsData[selectedChar].FolDis = USPF.ptsData.FolDis
	end
end

local function USPF_SetAllianceWarRankPoints()
	USPF.ptsData.PvPRank = GetUnitAvARank("player") == nil and 0 or GetUnitAvARank("player")

	--Update saved variables.
	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		USPF.sVar.ptsData[selectedChar].PvPRank = USPF.ptsData.PvPRank
	end
end

local function USPF_SetMaelArPoints()
	USPF.ptsData.MaelAr = IAchC(USPF.data.MAAch) and 1 or 0

	--Update saved variables.
	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		USPF.sVar.ptsData[selectedChar].MaelAr = USPF.ptsData.MaelAr
	end
end

local function USPF_SetUnassigned()
	USPF.ptsData.Unassigned = GetAvailableSkillPoints()

	--Update saved variables.
	if(USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar])) then
		USPF.sVar.ptsData[selectedChar].Unassigned = USPF.ptsData.Unassigned
	end
end

local function USPF_SetGeneralPoints()
	USPF.ptsData.GenTot =	USPF.ptsData.Level + USPF.ptsData.MainQ + USPF.ptsData.FolDis + USPF.ptsData.MWChar +
							USPF.ptsData.SUChar + USPF.ptsData.EWChar + USPF.ptsData.GMChar + USPF.ptsData.BWChar +
							USPF.ptsData.PvPRank + USPF.ptsData.MaelAr

	--Update saved variables.
	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		USPF.sVar.ptsData[selectedChar].GenTot = USPF.ptsData.GenTot
	end
end

local function USPF_SetTotPoints()
	USPF.ptsData.Tot =	USPF.ptsData.GenTot + USPF.ptsData.ZQTot + USPF.ptsData.SSTot +
						USPF.ptsData.GDTot + USPF.ptsData.PDTot

	--Update saved variables.
	if USPF_CheckSavedVars(USPF.sVar.ptsData[selectedChar]) then
		USPF.sVar.ptsData[selectedChar].Tot = USPF.ptsData.Tot
	end
end

local function USPF_LoadData(charId)
	local sVarPtsData = USPF.sVar.ptsData[charId]

	if USPF.sVar.ptsData[charId] == nil then
		--Write the character settings table.
		USPF.sVar.settings[charId] = {}
		USPF.sVar.settings[charId] = USPF_LTF:CopyTable(USPF.settings)

		--Write the character points data table.
		USPF.sVar.ptsData[charId] = {}
		USPF.sVar.ptsData[charId] = USPF_LTF:CopyTable(USPF.ptsData)
	end
	USPF_UpdateGUITable(sVarPtsData)
end

local function USPF_SetupData(charId)
	if charId == GCCId() then
		--Reset All Points to Zero.
		USPF.ptsData = USPF_LTF:SimpleResetTable(USPF.ptsData, 0)

		--Update Level Points.
		USPF_SetLevelPoints()

		--Update Quest Points.
		USPF_SetQuestPoints()

		--Update Public Dungeon Points.
		USPF_SetPublicDungeonPoints()

		--Update Skyshard Points.
		USPF_SetSkyshardPoints()

		--Update Alliance War Rank Points.
		USPF_SetAllianceWarRankPoints()

		--Update MaelAr Arena Point.
		USPF_SetMaelArPoints()

		--Update Unassigned Skillpoints
		USPF_SetUnassigned()

		--Update Folium Discognitum Points.
		USPF_SetFoliumDiscognitumPoints()

		--Update General Points.
		USPF_SetGeneralPoints()

		--Update Tot Points.
		USPF_SetTotPoints()

		--Update the GUI Table.
		USPF_UpdateGUITable(USPF.ptsData)

		--Update the saved variables.
		USPF_UpdateAllSavedVars()
	else
		--Load selected character data.
		USPF_LoadData(charId)
	end
end

local function USPF_FormatProgress(current, total, colors)
	if total == 0 then return "-" end
	local color
	if current == 0 then
		color = colors.need
	elseif current == total then
		color = colors.done
	else
		color = colors.progress
	end
	return strF("%s%d/%d|r", color, current, total)
end

function USPF:UpdateDataLines()
	local dataLines_GSP, dataLines_SQS, dataLines_GDQ, dataLines_PDGBE= {},{},{},{}

	USPF_SetupData(selectedChar)

	local GSP_Color = {
		need = USPF_rgbToHex(USPF.settings.GSP.needColor),
		progress = USPF_rgbToHex(USPF.settings.GSP.progColor),
		done = USPF_rgbToHex(USPF.settings.GSP.doneColor)
	}
	local SQS_ColorZQ = {
		need =USPF_rgbToHex(USPF.settings.SQS.needColorZQ),
		progress = USPF_rgbToHex(USPF.settings.SQS.progColorZQ),
		done = USPF_rgbToHex(USPF.settings.SQS.doneColorZQ)
	}
	local SQS_ColorSS = {
		need = USPF_rgbToHex(USPF.settings.SQS.needColorSS),
		progress = USPF_rgbToHex(USPF.settings.SQS.progColorSS),
		done = USPF_rgbToHex(USPF.settings.SQS.doneColorSS)
	}
	local GDQ_Color = {
		need = USPF_rgbToHex(USPF.settings.GDQ.needColor),
		done = USPF_rgbToHex(USPF.settings.GDQ.doneColor)
	}
	local PDB_Color = {
		need = USPF_rgbToHex(USPF.settings.PDB.needColor),
		done = USPF_rgbToHex(USPF.settings.PDB.doneColor)
	}

	for i = 1, #USPF.GUI.GSP do
		table.insert(dataLines_GSP, {
			source = USPF.GUI.GSP[i][2],
			progress = USPF_FormatProgress(USPF.GUI.GSP[i][3], USPF.GUI.GSP[i][4], GSP_Color)
		})
	end

	local tempTable = USPF_LTF:SortTable(USPF.GUI.SQS, USPF.settings.SQS.sortCol)
	for i = 1, #tempTable do
		table.insert(dataLines_SQS, {
			zone = tempTable[i][2],
			quests = USPF_FormatProgress(tempTable[i][3], tempTable[i][4], SQS_ColorZQ),
			skyshards = USPF_FormatProgress(tempTable[i][5], tempTable[i][6], SQS_ColorSS),
		})
	end

	tempTable = USPF_LTF:SortTable(USPF.GUI.GDQ, USPF.settings.GDQ.sortCol)
	for i = 1, #tempTable do
		table.insert(dataLines_GDQ, {
			zone = tempTable[i][2],
			dungeon = tempTable[i][3],
			progress = USPF_FormatProgress(tempTable[i][4], tempTable[i][5], GDQ_Color),
		})
	end

	tempTable = USPF_LTF:SortTable(USPF.GUI.PDGBE, USPF.settings.PDB.sortCol)
	for i = 1, #tempTable do
		table.insert(dataLines_PDGBE, {
			zone = tempTable[i][2],
			dungeon = tempTable[i][3],
			progress = USPF_FormatProgress(tempTable[i][4], tempTable[i][5], PDB_Color),
		})
	end


	USPF_GUI_Header_Title:SetText(GS(USPF_GUI_TITLE))
	USPF_GUI_Body_GSP:SetText(GS(USPF_GUI_GSP))
	USPF_GUI_Body_GSP_Source:SetText(GS(USPF_GUI_SOURCE))
	USPF_GUI_Body_GSP_Progress:SetText(GS(USPF_GUI_PROGRESS))

	USPF_GUI_Body_SQS:SetText(GS(USPF_GUI_SQS))
	USPF_GUI_Body_SQS_Z:SetText(GS(USPF_GUI_ZONE))
	USPF_GUI_Body_SQS_SL:SetText(GS(USPF_GUI_STORYLINE))
	USPF_GUI_Body_SQS_SS:SetText(GS(USPF_GUI_SKYSHARDS))
	USPF_GUI_Body_SQS_Z_T:SetText(GS(USPF_GUI_TOTAL)..":")
	USPF_GUI_Body_SQS_Z_T.data = {tooltipText = getTooltipZoneTotal()}
	USPF_GUI_Body_SQS_Z_T:SetMouseEnabled(true)
	USPF_GUI_Body_SQS_Z_T:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
	USPF_GUI_Body_SQS_Z_T:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)

	USPF_GUI_Body_GDQ:SetText(GS(USPF_GUI_GDQ))
	USPF_GUI_Body_GDQ_Z:SetText(GS(USPF_GUI_ZONE))
	USPF_GUI_Body_GDQ_D:SetText(GS(USPF_GUI_GROUP_DUNGEON))
	USPF_GUI_Body_GDQ_P:SetText(GS(USPF_GUI_PROGRESS))

	USPF_GUI_Body_PDGBE:SetText(GS(USPF_GUI_PDB))
	USPF_GUI_Body_PDGBE_Z:SetText(GS(USPF_GUI_ZONE))
	USPF_GUI_Body_PDGBE_D:SetText(GS(USPF_GUI_PUBLIC_DUNGEON))
	USPF_GUI_Body_PDGBE_P:SetText(GS(USPF_GUI_PROGRESS))


	USPF_GUI_Body_GSP_ListHolder.dataLines = dataLines_GSP
	USPF_GUI_Body_GSP_ListHolder:SetParent(USPF_GUI_Body)
	USPF_GUI_Body_SQS_ListHolder.dataLines = dataLines_SQS
	USPF_GUI_Body_SQS_ListHolder:SetParent(USPF_GUI_Body)
	USPF_GUI_Body_GDQ_ListHolder.dataLines = dataLines_GDQ
	USPF_GUI_Body_GDQ_ListHolder:SetParent(USPF_GUI_Body)
	USPF_GUI_Body_PDGBE_ListHolder.dataLines = dataLines_PDGBE
	USPF_GUI_Body_PDGBE_ListHolder:SetParent(USPF_GUI_Body)

	USPF_GUI_Body_GSP_T:SetText(USPF.GUI.GSP_T)

	USPF_GUI_Body_SQS_SL_T:SetText(USPF.GUI.SQS_SL_T)
	USPF_GUI_Body_SQS_SL_T.data = {tooltipText = getTooltipZoneTotal()}
	USPF_GUI_Body_SQS_SL_T:SetMouseEnabled(true)
	USPF_GUI_Body_SQS_SL_T:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
	USPF_GUI_Body_SQS_SL_T:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)

	USPF_GUI_Body_SQS_SS_T:SetText(USPF.GUI.SQS_SS_T)
	USPF_GUI_Body_SQS_SS_T.data = {tooltipText = getTooltipZoneTotal()}
	USPF_GUI_Body_SQS_SS_T:SetMouseEnabled(true)
	USPF_GUI_Body_SQS_SS_T:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
	USPF_GUI_Body_SQS_SS_T:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)

	USPF_GUI_Body_GDQ_T:SetText(USPF.GUI.GDQ_T)
	USPF_GUI_Body_GDQ_T.data = {tooltipText = getTooltipGDTotal()}
	USPF_GUI_Body_GDQ_T:SetMouseEnabled(true)
	USPF_GUI_Body_GDQ_T:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
	USPF_GUI_Body_GDQ_T:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)

	USPF_GUI_Body_PDGBE_T:SetText(USPF.GUI.PDGBE_T)
	USPF_GUI_Body_PDGBE_T.data = {tooltipText = getTooltipPDTotal()}
	USPF_GUI_Body_PDGBE_T:SetMouseEnabled(true)
	USPF_GUI_Body_PDGBE_T:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
	USPF_GUI_Body_PDGBE_T:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)

	USPF_GUI_Footer_CharacterTotal:SetText(USPF.GUI.CharacterTot)
	USPF_GUI_Footer_CharacterTotal.data = {tooltipText = getTooltipCharacterTotal()}
	USPF_GUI_Footer_CharacterTotal:SetMouseEnabled(true)
	USPF_GUI_Footer_CharacterTotal:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
	USPF_GUI_Footer_CharacterTotal:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)
	USPF:InitializeDataLines()
end

function USPF:FillLine(currLine, currItem)
	local cln, clsrc, clz, clq, clss, cld, clp = currLine:GetParent():GetName(), currLine.source, currLine.zone, currLine.quests, currLine.skyshards, currLine.dungeon, currLine.progress

	if currItem == nil then
		if cln == "USPF_GUI_Body_GSP_ListHolder" then
			clsrc:SetText("")
			clp:SetText("")
		elseif cln == "USPF_GUI_Body_SQS_ListHolder" then
			clz:SetText("")
			clq:SetText("")
			clss:SetText("")
		elseif cln == "USPF_GUI_Body_GDQ_ListHolder" then
			clz:SetText("")
			cld:SetText("")
			clp:SetText("")
		elseif cln == "USPF_GUI_Body_PDGBE_ListHolder" then
			clz:SetText("")
			cld:SetText("")
			clp:SetText("")
		end
	else
		if cln == "USPF_GUI_Body_GSP_ListHolder" then
			clsrc:SetText(currItem.source)
			clp:SetText(currItem.progress)
		elseif cln == "USPF_GUI_Body_SQS_ListHolder" then
			clz:SetText(currItem.zone)
			clq:SetText(currItem.quests)
			clss:SetText(currItem.skyshards)
		elseif cln == "USPF_GUI_Body_GDQ_ListHolder" then
			clz:SetText(currItem.zone)
			cld:SetText(currItem.dungeon)
			clp:SetText(currItem.progress)
		elseif cln == "USPF_GUI_Body_PDGBE_ListHolder" then
			clz:SetText(currItem.zone)
			cld:SetText(currItem.dungeon)
			clp:SetText(currItem.progress)
		end
	end
end

function USPF:InitializeDataLines()
	local currLine, currData
	for i = 1, #USPF.GUI.GSP do
		currLine = USPF_GUI_Body_GSP_ListHolder.lines[i]
		currData = USPF_GUI_Body_GSP_ListHolder.dataLines[i]
		USPF:FillLine(currLine, currData)

		if USPF.GUI.GSP[i][5] ~= nil then
			-- tooltip text
			USPF_GUI_Body_GSP_ListHolder.lines[i].data = {tooltipText = USPF.GUI.GSP[i][5]}

			-- tooltip handlers
			USPF_GUI_Body_GSP_ListHolder.lines[i]:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
			USPF_GUI_Body_GSP_ListHolder.lines[i]:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)
		end
	end

	for i = 1, #USPF.GUI.SQS do
		currLine = USPF_GUI_Body_SQS_ListHolder.lines[i]
		currData = USPF_GUI_Body_SQS_ListHolder.dataLines[i]
		USPF:FillLine(currLine, currData)

		if USPF.GUI.SQS[i][7] ~= nil then
			-- tooltip text
			USPF_GUI_Body_SQS_ListHolder.lines[i].data = {tooltipText = USPF.GUI.SQS[i][7]}

			-- tooltip handlers
			USPF_GUI_Body_SQS_ListHolder.lines[i]:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
			USPF_GUI_Body_SQS_ListHolder.lines[i]:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)
		end
	end

	for i = 1, #USPF.GUI.GDQ do
		currLine = USPF_GUI_Body_GDQ_ListHolder.lines[i]
		currData = USPF_GUI_Body_GDQ_ListHolder.dataLines[i]
		USPF:FillLine(currLine, currData)

		if USPF.GUI.GDQ[i][6] ~= nil then
			-- tooltip text
			USPF_GUI_Body_GDQ_ListHolder.lines[i].data = {tooltipText = USPF.GUI.GDQ[i][6]}

			-- tooltip handlers
			USPF_GUI_Body_GDQ_ListHolder.lines[i]:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
			USPF_GUI_Body_GDQ_ListHolder.lines[i]:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)
		end
	end

	for i = 1, #USPF.GUI.PDGBE do
		currLine = USPF_GUI_Body_PDGBE_ListHolder.lines[i]
		currData = USPF_GUI_Body_PDGBE_ListHolder.dataLines[i]
		USPF:FillLine(currLine, currData)

		if USPF.GUI.PDGBE[i][6] ~= nil then
			-- tooltip text
			USPF_GUI_Body_PDGBE_ListHolder.lines[i].data = {tooltipText = USPF.GUI.PDGBE[i][6]}

			-- tooltip handlers
			USPF_GUI_Body_PDGBE_ListHolder.lines[i]:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
			USPF_GUI_Body_PDGBE_ListHolder.lines[i]:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)
		end
	end
end

function USPF:ToggleWindow()
	USPF.active = not USPF.active
	if USPF.active then USPF:UpdateDataLines() end
	USPF_GUI:SetHidden(not USPF.active)
end


function USPF:CreateLine(i, USPF_predecessor, USPF_parent)
	if USPF_parent:GetName() == "USPF_GUI_Body_GSP_ListHolder" then
		local USPF_record = CreateControlFromVirtual("USPF_GSP_Row_", USPF_parent, "USPF_GeneralTemplate", i)

		USPF_record.source = USPF_record:GetNamedChild("_Source")
		USPF_record.progress = USPF_record:GetNamedChild("_Progress")

		USPF_record.source:SetFont(USPF.Options.Font.Fonts[USPF.settings.GSP.font].."|14")
		USPF_record.progress:SetFont(USPF.Options.Font.Fonts[USPF.settings.GSP.font].."|14")

		USPF_record:SetHidden(false)
		USPF_record:SetMouseEnabled(true)
		USPF_record:SetHeight("18")

		if i == 1 then
			USPF_record:SetAnchor(TOPLEFT, USPF_GUI_Body_GSP_ListHolder, TOPLEFT, 0, 0)
			USPF_record:SetAnchor(TOPRIGHT, USPF_GUI_Body_GSP_ListHolder, TOPRIGHT, 0, 0)
		else
			USPF_record:SetAnchor(TOPLEFT, USPF_predecessor, BOTTOMLEFT, 0, USPF_GUI_Body_GSP_ListHolder.rowHeight)
			USPF_record:SetAnchor(TOPRIGHT, USPF_predecessor, BOTTOMRIGHT, 0, USPF_GUI_Body_GSP_ListHolder.rowHeight)
		end
		USPF_record:SetParent(USPF_GUI_Body_GSP_ListHolder)
		return USPF_record
	elseif USPF_parent:GetName() == "USPF_GUI_Body_SQS_ListHolder" then
		local USPF_record = CreateControlFromVirtual("USPF_SQS_Row_", USPF_parent, "USPF_SQSSTemplate", i)

		USPF_record.zone = USPF_record:GetNamedChild("_Zone")
		USPF_record.quests = USPF_record:GetNamedChild("_Quests")
		USPF_record.skyshards = USPF_record:GetNamedChild("_Skyshards")

		USPF_record.zone:SetFont(USPF.Options.Font.Fonts[USPF.settings.SQS.font].."|14")
		USPF_record.quests:SetFont(USPF.Options.Font.Fonts[USPF.settings.SQS.font].."|14")
		USPF_record.skyshards:SetFont(USPF.Options.Font.Fonts[USPF.settings.SQS.font].."|14")

		USPF_record:SetHidden(false)
		USPF_record:SetMouseEnabled(true)
		USPF_record:SetHeight("18")

		if i == 1 then
			USPF_record:SetAnchor(TOPLEFT, USPF_GUI_Body_SQS_ListHolder, TOPLEFT, 0, 0)
			USPF_record:SetAnchor(TOPRIGHT, USPF_GUI_Body_SQS_ListHolder, TOPRIGHT, 0, 0)
		else
			USPF_record:SetAnchor(TOPLEFT, USPF_predecessor, BOTTOMLEFT, 0, USPF_GUI_Body_SQS_ListHolder.rowHeight)
			USPF_record:SetAnchor(TOPRIGHT, USPF_predecessor, BOTTOMRIGHT, 0, USPF_GUI_Body_SQS_ListHolder.rowHeight)
		end
		USPF_record:SetParent(USPF_GUI_Body_SQS_ListHolder)
		return USPF_record
	elseif USPF_parent:GetName() == "USPF_GUI_Body_GDQ_ListHolder" then
		local USPF_record = CreateControlFromVirtual("USPF_GDQ_Row_", USPF_parent, "USPF_GDQTemplate", i)

		USPF_record.zone = USPF_record:GetNamedChild("_Zone")
		USPF_record.dungeon = USPF_record:GetNamedChild("_Dungeon")
		USPF_record.progress = USPF_record:GetNamedChild("_Progress")

		USPF_record.zone:SetFont(USPF.Options.Font.Fonts[USPF.settings.GDQ.font].."|14")
		USPF_record.dungeon:SetFont(USPF.Options.Font.Fonts[USPF.settings.GDQ.font].."|14")
		USPF_record.progress:SetFont(USPF.Options.Font.Fonts[USPF.settings.GDQ.font].."|14")

		USPF_record:SetHidden(false)
		USPF_record:SetMouseEnabled(true)
		USPF_record:SetHeight("18")

		if i == 1 then
			USPF_record:SetAnchor(TOPLEFT, USPF_GUI_Body_GDQ_ListHolder, TOPLEFT, 0, 0)
			USPF_record:SetAnchor(TOPRIGHT, USPF_GUI_Body_GDQ_ListHolder, TOPRIGHT, 0, 0)
		else
			USPF_record:SetAnchor(TOPLEFT, USPF_predecessor, BOTTOMLEFT, 0, USPF_GUI_Body_GDQ_ListHolder.rowHeight)
			USPF_record:SetAnchor(TOPRIGHT, USPF_predecessor, BOTTOMRIGHT, 0, USPF_GUI_Body_GDQ_ListHolder.rowHeight)
		end
		USPF_record:SetParent(USPF_GUI_Body_GDQ_ListHolder)
		return USPF_record
	elseif USPF_parent:GetName() == "USPF_GUI_Body_PDGBE_ListHolder" then
		local USPF_record = CreateControlFromVirtual("USPF_PDGBE_Row_", USPF_parent, "USPF_PDGBETemplate", i)

		USPF_record.zone = USPF_record:GetNamedChild("_Zone")
		USPF_record.dungeon = USPF_record:GetNamedChild("_Dungeon")
		USPF_record.progress = USPF_record:GetNamedChild("_Progress")

		USPF_record.zone:SetFont(USPF.Options.Font.Fonts[USPF.settings.PDB.font].."|14")
		USPF_record.dungeon:SetFont(USPF.Options.Font.Fonts[USPF.settings.PDB.font].."|14")
		USPF_record.progress:SetFont(USPF.Options.Font.Fonts[USPF.settings.PDB.font].."|14")

		USPF_record:SetHidden(false)
		USPF_record:SetMouseEnabled(true)
		USPF_record:SetHeight("18")

		if i == 1 then
			USPF_record:SetAnchor(TOPLEFT, USPF_GUI_Body_PDGBE_ListHolder, TOPLEFT, 0, 0)
			USPF_record:SetAnchor(TOPRIGHT, USPF_GUI_Body_PDGBE_ListHolder, TOPRIGHT, 0, 0)
		else
			USPF_record:SetAnchor(TOPLEFT, USPF_predecessor, BOTTOMLEFT, 0, USPF_GUI_Body_PDGBE_ListHolder.rowHeight)
			USPF_record:SetAnchor(TOPRIGHT, USPF_predecessor, BOTTOMRIGHT, 0, USPF_GUI_Body_PDGBE_ListHolder.rowHeight)
		end
		USPF_record:SetParent(USPF_GUI_Body_PDGBE_ListHolder)
		return USPF_record
	end
end

function USPF:CreateListHolders()
	local USPF_predecessor = nil
	USPF_GUI_Body_GSP_ListHolder.dataLines = {}
	USPF_GUI_Body_GSP_ListHolder.lines = {}
	for i = 1, #USPF.GUI.GSP do
		USPF_GUI_Body_GSP_ListHolder.lines[i] = USPF:CreateLine(i, USPF_predecessor, USPF_GUI_Body_GSP_ListHolder)
		USPF_predecessor = USPF_GUI_Body_GSP_ListHolder.lines[i]
	end

	USPF_predecessor = nil
	USPF_GUI_Body_SQS_ListHolder.dataLines = {}
	USPF_GUI_Body_SQS_ListHolder.lines = {}
	for i = 1, #USPF.GUI.SQS do
		USPF_GUI_Body_SQS_ListHolder.lines[i] = USPF:CreateLine(i, USPF_predecessor, USPF_GUI_Body_SQS_ListHolder)
		USPF_predecessor = USPF_GUI_Body_SQS_ListHolder.lines[i]
	end

	USPF_predecessor = nil
	USPF_GUI_Body_GDQ_ListHolder.dataLines = {}
	USPF_GUI_Body_GDQ_ListHolder.lines = {}
	for i = 1, #USPF.GUI.GDQ do
		USPF_GUI_Body_GDQ_ListHolder.lines[i] = USPF:CreateLine(i, USPF_predecessor, USPF_GUI_Body_GDQ_ListHolder)
		USPF_predecessor = USPF_GUI_Body_GDQ_ListHolder.lines[i]
	end

	USPF_predecessor = nil
	USPF_GUI_Body_PDGBE_ListHolder.dataLines = {}
	USPF_GUI_Body_PDGBE_ListHolder.lines = {}
	for i = 1, #USPF.GUI.PDGBE do
		USPF_GUI_Body_PDGBE_ListHolder.lines[i] = USPF:CreateLine(i, USPF_predecessor, USPF_GUI_Body_PDGBE_ListHolder)
		USPF_predecessor = USPF_GUI_Body_PDGBE_ListHolder.lines[i]
	end

	USPF:UpdateDataLines()
end

function USPF:SetupValues()
	--Reset All Points to Zero.
	USPF.ptsData = USPF_LTF:SimpleResetTable(USPF.ptsData, 0)

	--Update Level Points.
	USPF_SetLevelPoints()

	--Update Quest Points.
	USPF_SetQuestPoints()

	--Update Public Dungeon Points.
	USPF_SetPublicDungeonPoints()

	--Update Skyshard Points.
	USPF_SetSkyshardPoints()

	--Update Alliance War Rank Points.
	USPF_SetAllianceWarRankPoints()

	--Update MaelAr Arena Point.
	USPF_SetMaelArPoints()

	--Update Folium Discognitum Points.
	USPF_SetFoliumDiscognitumPoints()

	--Update General Points.
	USPF_SetGeneralPoints()

	--Update Tot Points.
	USPF_SetTotPoints()

	--Update the GUI Table.
	USPF_UpdateGUITable(USPF.ptsData)

	--Set the window size and position.
	USPF_GUI:ClearAnchors()
	USPF_GUI:SetAnchor(CENTER, GuiRoot, CENTER, 0, 0)
	USPF_GUI:SetHeight(USPF_GUI_Header:GetHeight() + #USPF.GUI.GDQ * 18 + USPF_GUI_Footer:GetHeight() + 18 + 76)

	--Add information to window.
	USPF:CreateListHolders()

	USPF_GUI_Header:SetParent(USPF_GUI)
	USPF_GUI_Body:SetParent(USPF_GUI)
	USPF_GUI_Footer:SetParent(USPF_GUI)

	USPF_GUI_Body_GSP_ListHolder:SetParent(USPF_GUI_Body)
	USPF_GUI_Body_SQS_ListHolder:SetParent(USPF_GUI_Body)
	USPF_GUI_Body_GDQ_ListHolder:SetParent(USPF_GUI_Body)
	USPF_GUI_Body_PDGBE_ListHolder:SetParent(USPF_GUI_Body)

	USPF_GUI_Header_Title:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|30")
	USPF_GUI_Body_GSP:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|16")
	USPF_GUI_Body_GSP_Source:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_GSP_Progress:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_GSP_T:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")

	USPF_GUI_Body_SQS:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|16")
	USPF_GUI_Body_SQS_Z:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_SQS_SL:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_SQS_SS:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_SQS_Z_T:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_SQS_SL_T:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_SQS_SS_T:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")

	USPF_GUI_Body_GDQ:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|16")
	USPF_GUI_Body_GDQ_Z:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_GDQ_D:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_GDQ_P:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_GDQ_T:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")

	USPF_GUI_Body_PDGBE:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|16")
	USPF_GUI_Body_PDGBE_Z:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_PDGBE_D:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_PDGBE_P:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")
	USPF_GUI_Body_PDGBE_T:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|14")

	USPF_GUI_Footer_CharacterTotal:SetFont(USPF.Options.Font.Fonts[USPF.settings.title.font].."|24")
end

function USPF:HelpSlash()
	d(GS(USPF_MSG_CMD_TITLE))
	d(strF(GS(USPF_MSG_CMD_OPTION), (USPF.active and GS(USPF_MSG_DEACTVATE) or GS(USPF_MSG_ACTIVATE))))
end

function USPF:BadSlash()
	d(GS(USPF_MSG_BAD_SLASH))
	d(GS(USPF_MSG_CMD_TITLE))
	d(strF(GS(USPF_MSG_CMD_OPTION), (USPF.active and GS(USPF_MSG_DEACTVATE) or GS(USPF_MSG_ACTIVATE))))
end

local function USPF_LoadSettings(charId)
	if USPF_CheckSavedVars(USPF.sVar.settings[charId]) then
		USPF.settings = USPF_LTF:CopyTable(USPF.sVar.settings[charId])

		if USPF.settings.GSP.progColor == nil then
			USPF.settings.GSP.progColor = {1, 1, 1}
			USPF.sVar.settings[charId].GSP.progColor = USPF.settings.GSP.progColor
		end

		if USPF.settings.SQS.progColorSS == nil then
			USPF.settings.SQS.progColorSS = {0.7843, 0.3922, 0}
			USPF.sVar.settings[charId].SQS.progColorSS = USPF.settings.SQS.progColorSS
		end

		if USPF.settings.SQS.progColorZQ == nil then
			USPF.settings.SQS.progColorZQ = {0.7843, 0.3922, 0}
			USPF.sVar.settings[charId].SQS.progColorZQ = USPF.settings.SQS.progColorZQ
		end

		USPF.settings.MWC = USPF.sVar.settings[charId].SSC and false or USPF.sVar.settings[charId].MWC
		USPF.settings.SSC = USPF.sVar.settings[charId].MWC and false or USPF.sVar.settings[charId].SSC
		USPF.sVar.settings[charId].MWC = USPF.settings.SSC and false or USPF.settings.MWC
		USPF.sVar.settings[charId].SSC = USPF.settings.MWC and false or USPF.settings.SSC
	end
end

local function USPF_SetSelectedChar(charName)
	for i = 1, #USPF.charData do
		if charName == USPF.charData[i].charName then
			selectedChar = USPF.charData[i].charId
			break
		end
	end
end

local function USPF_CreateCharList()
	USPF_GUI_Header_CharList.comboBox = USPF_GUI_Header_CharList.comboBox or ZO_ComboBox_ObjectFromContainer(USPF_GUI_Header_CharList)
	local USPF_comboBox = USPF_GUI_Header_CharList.comboBox

	-- tooltip text
	USPF_GUI_Header_CharList.data = {tooltipText = "Select a character to view. You must have logged in on the selected character with this addon active for information to populate."}

	-- tooltip handlers
	USPF_GUI_Header_CharList:SetHandler("OnMouseEnter", ZO_Options_OnMouseEnter)
	USPF_GUI_Header_CharList:SetHandler("OnMouseExit", ZO_Options_OnMouseExit)

	USPF.charNames = {}
	USPF.charData = {}
	for k,_ in ipairs(USPF.sVar.charInfo) do
		USPF.charNames[k] = USPF.sVar.charInfo[k].charName
		USPF.charData[k] = {
			charId = USPF.sVar.charInfo[k].charId,
			charName = USPF.sVar.charInfo[k].charName,
		}

		if GCCId() == USPF.charData[k].charId then
			currentCharName = USPF.charData[k].charName
			selectedChar = USPF.charData[k].charId
		end
	end

	function OnItemSelect(_, choiceText, choice)
		USPF_SetSelectedChar(choiceText)
		USPF:UpdateDataLines()
		PlaySound(SOUNDS.POSITIVE_CLICK)
	end

	USPF_comboBox:SetSortsItems(false)

	for k,_ in ipairs(USPF.charNames) do
		USPF_comboBox:AddItem(USPF_comboBox:CreateItemEntry(USPF.charNames[k], OnItemSelect))
		if USPF.charNames[k] == currentCharName then
			USPF_comboBox:SetSelectedItem(USPF.charNames[k])
		end
	end
end

local function USPF_UpdateCharList()
	USPF.sVar.charInfo = {}
	for i = 1, GetNumCharacters() do
		local name, _, _, _, _, _, id, _ = GetCharacterInfo(i)
		USPF.sVar.charInfo[i] = {charId = id, charName = zf("<<t:1>>", name),}
	end
end

local function USPF_InitSetup()
	local firstRun = USPF.sVar.firstRun or USPF.sVar.numChars == 0 or USPF.sVar.charInfo[1] == nil

	--If first run then setup the character table
	if firstRun then
		d(GS(USPF_MSG_INIT))
		USPF.sVar.numChars = GetNumCharacters()
		for i = 1, GetNumCharacters() do
			local name, _, _, _, _, _, id, _ = GetCharacterInfo(i)
			USPF.sVar.charInfo[i] = {charId = id, charName = zf("<<t:1>>", name),}

			--Write the character settings table.
			USPF.sVar.settings[id] = {}
			USPF.sVar.settings[id] = USPF_LTF:CopyTable(USPF.settings)

			--Write the character points data table.
			USPF.sVar.ptsData[id] = {}
			USPF.sVar.ptsData[id] = USPF_LTF:CopyTable(USPF.ptsData)
		end

		--Show the welcome message.
		d(GS(USPF_MSG_HELP))

		USPF.sVar.firstRun = false
	end

	--Setup the character info table.
	if not firstRun then USPF_UpdateCharList() end

	--Check for added characters.
	for _,v in pairs(USPF.sVar.charInfo) do
		if not USPF_LTF:TableContains(USPF.sVar.settings, v.charId, true) then
			USPF.sVar.settings[v.charId] = USPF_LTF:CopyTable(USPF.settings)
		end

		if not USPF_LTF:TableContains(USPF.sVar.ptsData, v.charId, true) then
			USPF.sVar.ptsData[v.charId] = USPF_LTF:CopyTable(USPF.ptsData)
		end
	end

	--Check for removed characters.
	for k,_ in pairs(USPF.sVar.settings) do
		USPF.sVar.settings[k] = USPF_LTF:TableContains(USPF.sVar.charInfo, k, false) and USPF.sVar.settings[k] or nil
	end
	for k,_ in pairs(USPF.sVar.ptsData) do
		USPF.sVar.ptsData[k] = USPF_LTF:TableContains(USPF.sVar.charInfo, k, false) and USPF.sVar.ptsData[k] or nil
	end

	--Create the character select box.
	USPF_CreateCharList()
end

local function USPF_ResetSelectedCharacter()
	local currentCharId = GCCId()
	if selectedChar ~= currentCharId then
		USPF_GUI_Header_CharList.comboBox = USPF_GUI_Header_CharList.comboBox or ZO_ComboBox_ObjectFromContainer(USPF_GUI_Header_CharList)
		local USPF_comboBox = USPF_GUI_Header_CharList.comboBox

		for k,_ in ipairs(USPF.charData) do
			if currentCharId == USPF.charData[k].charId then
				currentCharName = USPF.charData[k].charName
				selectedChar = USPF.charData[k].charId
				USPF_comboBox:SetSelectedItem(currentCharName)
			end
		end
	end
	USPF_SetupData(currentCharId)
end


local function USPF_SkillPointsUpdate(eventCode, pointsBefore, pointsNow, partialPointsBefore, partialPointsNow)
	USPF_ResetSelectedCharacter()
end

local function USPF_QuestRemoved(eventCode, isCompleted, journalIndex, questName, zoneIndex, poiIndex, questID)
	if isCompleted then USPF_ResetSelectedCharacter() end
end

local function USPF_LevelUp(eventCode, unitTag, level)
	if unitTag == 'player' then USPF_ResetSelectedCharacter() end
end

local function USPF_AchComplete(eventCode, name, points, id, link)
	USPF_ResetSelectedCharacter()
end

local function USPF_PlayerActivated(eventCode)
	USPF_ResetSelectedCharacter()
end

local function USPF_PlayerDeactivated(eventCode)
	USPF_ResetSelectedCharacter()
end

local function USPF_Initialized(eventCode, addonName)
	if addonName ~= USPF.AddonName then return end

	local world = GetWorldName()
	if world == "NA Megaserver" then
		USPF.sVar = ZO_SavedVars:NewAccountWide("USPF_Settings", USPF.version, nil, USPF.defaults)
	else
		USPF.sVar = ZO_SavedVars:NewAccountWide("USPF_Settings", USPF.version, world, USPF.defaults)
	end

	--Run the startup routine.
	USPF_InitSetup()

	--Create the keybind(s).
	ZO_CreateStringId("SI_BINDING_NAME_USPF_TOGGLE", "Show Skill Point Finder")

	--Register the slash commands.
	SLASH_COMMANDS["/uspf"] = function(keyWord, argument)
		if string.lower(keyWord) == "help" then
			USPF:HelpSlash()
		elseif keyWord == "" then
			USPF:ToggleWindow()
		else
			USPF:BadSlash()
		end
	end

	local charId = GCCId()

	--Load the character settings.
	USPF_LoadSettings(charId)

	--Create the options menu.
	USPF:SetupMenu(charId)

	--Call startup routine.
	USPF:SetupValues()

	--Create the event handlers.
	EVENT_MANAGER:RegisterForEvent(USPF.AddonName, EVENT_SKILL_POINTS_CHANGED, USPF_SkillPointsUpdate)
	EVENT_MANAGER:RegisterForEvent(USPF.AddonName, EVENT_QUEST_REMOVED, USPF_QuestRemoved)
	EVENT_MANAGER:RegisterForEvent(USPF.AddonName, EVENT_LEVEL_UPDATE, USPF_LevelUp)
	EVENT_MANAGER:RegisterForEvent(USPF.AddonName, EVENT_ACHIEVEMENT_AWARDED, USPF_AchComplete)
	EVENT_MANAGER:RegisterForEvent(USPF.AddonName, EVENT_PLAYER_ACTIVATED, USPF_PlayerActivated)
	EVENT_MANAGER:RegisterForEvent(USPF.AddonName, EVENT_PLAYER_DEACTIVATED, USPF_PlayerDeactivated)

	--Kill the initial event handler.
	EVENT_MANAGER:UnregisterForEvent(USPF.AddonName, EVENT_ADD_ON_LOADED)
end

--Register initialization event.
EVENT_MANAGER:RegisterForEvent(USPF.AddonName, EVENT_ADD_ON_LOADED, USPF_Initialized)
